classnames = [
    "abbey",
    "airplane cabin",
    "airport terminal",
    "alley",
    "amphitheater",
    "amusement arcade",
    "amusement park",
    "anechoic chamber",
    "apartment building outdoor",
    "apse indoor",
    "aquarium",
    "aqueduct",
    "arch",
    "archive",
    "arrival gate outdoor",
    "art gallery",
    "art school",
    "art studio",
    "assembly line",
    "athletic field outdoor",
    "atrium public",
    "attic",
    "auditorium",
    "auto factory",
    "badlands",
    "badminton court indoor",
    "baggage claim",
    "bakery shop",
    "balcony exterior",
    "balcony interior",
    "ball pit",
    "ballroom",
    "bamboo forest",
    "banquet hall",
    "bar",
    "barn",
    "barndoor",
    "baseball field",
    "basement",
    "basilica",
    "basketball court outdoor",
    "bathroom",
    "batters box",
    "bayou",
    "bazaar indoor",
    "bazaar outdoor",
    "beach",
    "beauty salon",
    "bedroom",
    "berth",
    "biology laboratory",
    "bistro indoor",
    "boardwalk",
    "boat deck",
    "boathouse",
    "bookstore",
    "booth indoor",
    "botanical garden",
    "bow window indoor",
    "bow window outdoor",
    "bowling alley",
    "boxing ring",
    "brewery indoor",
    "bridge",
    "building facade",
    "bullring",
    "burial chamber",
    "bus interior",
    "butchers shop",
    "butte",
    "cabin outdoor",
    "cafeteria",
    "campsite",
    "campus",
    "canal natural",
    "canal urban",
    "candy store",
    "canyon",
    "car interior backseat",
    "car interior frontseat",
    "carrousel",
    "casino indoor",
    "castle",
    "catacomb",
    "cathedral indoor",
    "cathedral outdoor",
    "cavern indoor",
    "cemetery",
    "chalet",
    "cheese factory",
    "chemistry lab",
    "chicken coop indoor",
    "chicken coop outdoor",
    "childs room",
    "church indoor",
    "church outdoor",
    "classroom",
    "clean room",
    "cliff",
    "cloister indoor",
    "closet",
    "clothing store",
    "coast",
    "cockpit",
    "coffee shop",
    "computer room",
    "conference center",
    "conference room",
    "construction site",
    "control room",
    "control tower outdoor",
    "corn field",
    "corral",
    "corridor",
    "cottage garden",
    "courthouse",
    "courtroom",
    "courtyard",
    "covered bridge exterior",
    "creek",
    "crevasse",
    "crosswalk",
    "cubicle office",
    "dam",
    "delicatessen",
    "dentists office",
    "desert sand",
    "desert vegetation",
    "diner indoor",
    "diner outdoor",
    "dinette home",
    "dinette vehicle",
    "dining car",
    "dining room",
    "discotheque",
    "dock",
    "doorway outdoor",
    "dorm room",
    "driveway",
    "driving range outdoor",
    "drugstore",
    "electrical substation",
    "elevator door",
    "elevator interior",
    "elevator shaft",
    "engine room",
    "escalator indoor",
    "excavation",
    "factory indoor",
    "fairway",
    "fastfood restaurant",
    "field cultivated",
    "field wild",
    "fire escape",
    "fire station",
    "firing range indoor",
    "fishpond",
    "florist shop indoor",
    "food court",
    "forest broadleaf",
    "forest needleleaf",
    "forest path",
    "forest road",
    "formal garden",
    "fountain",
    "galley",
    "game room",
    "garage indoor",
    "garbage dump",
    "gas station",
    "gazebo exterior",
    "general store indoor",
    "general store outdoor",
    "gift shop",
    "golf course",
    "greenhouse indoor",
    "greenhouse outdoor",
    "gymnasium indoor",
    "hangar indoor",
    "hangar outdoor",
    "harbor",
    "hayfield",
    "heliport",
    "herb garden",
    "highway",
    "hill",
    "home office",
    "hospital",
    "hospital room",
    "hot spring",
    "hot tub outdoor",
    "hotel outdoor",
    "hotel room",
    "house",
    "hunting lodge outdoor",
    "ice cream parlor",
    "ice floe",
    "ice shelf",
    "ice skating rink indoor",
    "ice skating rink outdoor",
    "iceberg",
    "igloo",
    "industrial area",
    "inn outdoor",
    "islet",
    "jacuzzi indoor",
    "jail cell",
    "jail indoor",
    "jewelry shop",
    "kasbah",
    "kennel indoor",
    "kennel outdoor",
    "kindergarden classroom",
    "kitchen",
    "kitchenette",
    "labyrinth outdoor",
    "lake natural",
    "landfill",
    "landing deck",
    "laundromat",
    "lecture room",
    "library indoor",
    "library outdoor",
    "lido deck outdoor",
    "lift bridge",
    "lighthouse",
    "limousine interior",
    "living room",
    "lobby",
    "lock chamber",
    "locker room",
    "mansion",
    "manufactured home",
    "market indoor",
    "market outdoor",
    "marsh",
    "martial arts gym",
    "mausoleum",
    "medina",
    "moat water",
    "monastery outdoor",
    "mosque indoor",
    "mosque outdoor",
    "motel",
    "mountain",
    "mountain snowy",
    "movie theater indoor",
    "museum indoor",
    "music store",
    "music studio",
    "nuclear power plant outdoor",
    "nursery",
    "oast house",
    "observatory outdoor",
    "ocean",
    "office",
    "office building",
    "oil refinery outdoor",
    "oilrig",
    "operating room",
    "orchard",
    "outhouse outdoor",
    "pagoda",
    "palace",
    "pantry",
    "park",
    "parking garage indoor",
    "parking garage outdoor",
    "parking lot",
    "parlor",
    "pasture",
    "patio",
    "pavilion",
    "pharmacy",
    "phone booth",
    "physics laboratory",
    "picnic area",
    "pilothouse indoor",
    "planetarium outdoor",
    "playground",
    "playroom",
    "plaza",
    "podium indoor",
    "podium outdoor",
    "pond",
    "poolroom establishment",
    "poolroom home",
    "power plant outdoor",
    "promenade deck",
    "pub indoor",
    "pulpit",
    "putting green",
    "racecourse",
    "raceway",
    "raft",
    "railroad track",
    "rainforest",
    "reception",
    "recreation room",
    "residential neighborhood",
    "restaurant",
    "restaurant kitchen",
    "restaurant patio",
    "rice paddy",
    "riding arena",
    "river",
    "rock arch",
    "rope bridge",
    "ruin",
    "runway",
    "sandbar",
    "sandbox",
    "sauna",
    "schoolhouse",
    "sea cliff",
    "server room",
    "shed",
    "shoe shop",
    "shopfront",
    "shopping mall indoor",
    "shower",
    "skatepark",
    "ski lodge",
    "ski resort",
    "ski slope",
    "sky",
    "skyscraper",
    "slum",
    "snowfield",
    "squash court",
    "stable",
    "stadium baseball",
    "stadium football",
    "stage indoor",
    "staircase",
    "street",
    "subway interior",
    "subway station platform",
    "supermarket",
    "sushi bar",
    "swamp",
    "swimming pool indoor",
    "swimming pool outdoor",
    "synagogue indoor",
    "synagogue outdoor",
    "television studio",
    "temple east asia",
    "temple south asia",
    "tennis court indoor",
    "tennis court outdoor",
    "tent outdoor",
    "theater indoor procenium",
    "theater indoor seats",
    "thriftshop",
    "throne room",
    "ticket booth",
    "toll plaza",
    "topiary garden",
    "tower",
    "toyshop",
    "track outdoor",
    "train railway",
    "train station platform",
    "tree farm",
    "tree house",
    "trench",
    "underwater coral reef",
    "utility room",
    "valley",
    "van interior",
    "vegetable garden",
    "veranda",
    "veterinarians office",
    "viaduct",
    "videostore",
    "village",
    "vineyard",
    "volcano",
    "volleyball court indoor",
    "volleyball court outdoor",
    "waiting room",
    "warehouse indoor",
    "water tower",
    "waterfall block",
    "waterfall fan",
    "waterfall plunge",
    "watering hole",
    "wave",
    "wet bar",
    "wheat field",
    "wind farm",
    "windmill",
    "wine cellar barrel storage",
    "wine cellar bottle storage",
    "wrestling ring indoor",
    "yard",
    "youth hostel",
]

templates = [
    lambda c: f"a photo of a {c}.",
    lambda c: f"a photo of the {c}.",
]



import functools
import logging
import os
from typing import TYPE_CHECKING, Any, List, Optional, TypeVar

import lightning as L
import torch
from lightning.fabric.loggers import TensorBoardLogger
from lightning.fabric.utilities.rank_zero import rank_zero_only
from omegaconf import DictConfig, OmegaConf

from fusion_bench.utils import import_object
from fusion_bench.utils.instantiate import instantiate

if TYPE_CHECKING:
    import lightning.fabric.loggers.tensorboard
    from lightning.fabric.strategies import FSDPStrategy

log = logging.getLogger(__name__)

TensorOrModule = TypeVar("TensorOrModule", torch.Tensor, torch.nn.Module, Any)


def get_policy(*args: str) -> set:
    """
    Get the policy from the provided list of policy names.

    Args:
        *args (str): A list of policy names.

    Returns:
        set: A set of policy objects.
    """
    return {import_object(arg) for arg in args}


def get_size_based_auto_wrap_policy(*args, **kwargs):
    from torch.distributed.fsdp.wrap import size_based_auto_wrap_policy

    policy = functools.partial(size_based_auto_wrap_policy, *args, **kwargs)
    return policy


class LightningFabricMixin:
    """
    A mixin class for integrating Lightning Fabric into a project.

    This class provides methods to initialize and manage a Lightning Fabric instance for distributed computing,
    including setup with optional logging, device management for tensors and modules, and hyperparameter logging.
    It leverages the Lightning framework to facilitate distributed training and inference across multiple devices
    and nodes, with support for custom logging via TensorBoard.

    Attributes:
    - _fabric (L.Fabric): The Lightning Fabric instance used for distributed computing.

    Note:
    This mixin is designed to be used with classes that require distributed computing capabilities and wish to
    leverage the Lightning Fabric for this purpose. It assumes the presence of a `config` attribute or parameter
    in the consuming class for configuration.
    """

    _fabric_instance: L.Fabric = None

    def setup_lightning_fabric(self, config: DictConfig):
        """
        Initializes and launches the Lightning Fabric with optional logging.

        This method sets up the Lightning Fabric for distributed computing based on the provided configuration. If a fabric
        configuration is not found, it logs a warning and exits. Optionally, if a fabric logger configuration is provided,
        it initializes a TensorBoardLogger with the specified settings.

        Expected configuration keys:
        - fabric: The configuration for the Lightning Fabric.
        - fabric.loggers: The configuration for the TensorBoardLogger.
        """
        if self._fabric_instance is None:
            if config.get("fabric", None) is None:
                log.warning("No fabric configuration found. use default settings.")
                self._fabric_instance = L.Fabric()
            else:
                self._fabric_instance = instantiate(config.fabric)
            self._fabric_instance.launch()
            # Set the log directory in config if it is not already set
            if (
                self.log_dir is not None
                and hasattr(config, "log_dir")
                and config.get("log_dir", None) is None
            ):
                if self._fabric_instance.is_global_zero:
                    log.info(f"Setting log_dir to {self.log_dir}")
                config.log_dir = self.log_dir

    @property
    def fabric(self):
        if self._fabric_instance is None:
            self.setup_lightning_fabric(getattr(self, "config", DictConfig({})))
        return self._fabric_instance

    @property
    def log_dir(self):
        """
        Retrieves the log directory from the fabric's logger.
        """
        if self.fabric is not None and len(self.fabric._loggers) > 0:
            log_dir = self.fabric.logger.log_dir
            if self.fabric.is_global_zero and not os.path.exists(log_dir):
                os.makedirs(log_dir, exist_ok=True)
            return log_dir
        else:
            return None

    def to_device(self, obj: TensorOrModule) -> TensorOrModule:
        """
        Moves a tensor or module to the proper device.

        Args:
            obj (TensorOrModule): The tensor or module to move to the device.

        Returns:
            TensorOrModule: the same type of object as the input, moved to the device.
        """
        return self.fabric.to_device(obj)

    @rank_zero_only
    def log_hyperparams(
        self,
        config: Optional[DictConfig] = None,
        save_dir: Optional[str] = None,
        filename: str = "config.yaml",
    ):
        R"""
        Logs the hyperparameters and saves the configuration to a YAML file.
        The YAML file is saved in the log directory by default with the name `config.yaml`, or in the specified save directory `save_dir`.

        Args:
            config (Optional[DictConfig]): The configuration to log and save. If not provided, the class's `config` attribute is used.
            save_dir (Optional[str]): The directory in which to save the configuration file. If not provided, the log directory is used.
            filename (str): The name of the configuration file. Default is `config.yaml`.
        """
        if config is None:
            config = self.config
        if save_dir is None:
            save_dir = self.log_dir
        self.fabric.logger.log_hyperparams(
            OmegaConf.to_container(config, resolve=True, enum_to_str=True)
        )
        if not os.path.exists(save_dir):
            os.makedirs(save_dir, exist_ok=True)
        OmegaConf.save(
            config,
            os.path.join(self.log_dir if save_dir is None else save_dir, filename),
        )

    @property
    def tensorboard_summarywriter(
        self,
    ) -> "lightning.fabric.loggers.tensorboard.SummaryWriter":
        if isinstance(self.fabric.logger, TensorBoardLogger):
            return self.fabric.logger.experiment
        else:
            raise AttributeError("the logger is not a TensorBoardLogger.")

    @property
    def is_debug_mode(self):
        if hasattr(self, "config") and self.config.get("fast_dev_run", False):
            return True
        elif hasattr(self, "_program") and self._program.config.get(
            "fast_dev_run", False
        ):
            return True
        else:
            return False



"""
Modified from pytorch_optimizer: https://github.com/kozistr/pytorch_optimizer/blob/main/pytorch_optimizer/lr_scheduler/linear_warmup.py
"""

import math
from abc import ABC, abstractmethod
from typing import List

import numpy as np
import torch

from fusion_bench.optim.exception import NegativeLRError, NegativeStepError

__all__ = [
    "BaseLinearWarmupScheduler",
    "LinearWarmupScheduler",
    "CosineDecayWithWarmup",
    "PolySchedulerWithWarmup",
]


class BaseLinearWarmupScheduler(ABC):
    r"""BaseLinearWarmupScheduler class.

    The LR Scheduler class based on this class has linear warmup strategy.

    Args:
        optimizer (torch.optim.Optimizer): Optimizer. It will set learning rate to all trainable parameters in optimizer.
        T_max (int): Total steps to train.
        max_lr (float): Maximum learning rate.
        min_lr (float): Minimum learning rate.
        init_lr (float): Initial learning rate.
        warmup_steps (int): Steps to warm-up.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        T_max: int,
        max_lr: float,
        min_lr: float = 0.0,
        init_lr: float = 0.0,
        warmup_steps: int = 0,
    ):
        """
        Initialize the BaseLinearWarmupScheduler.

        Args:
            optimizer (torch.optim.Optimizer): Optimizer to apply the learning rate schedule.
            T_max (int): Total number of training steps.
            max_lr (float): Maximum learning rate.
            min_lr (float): Minimum learning rate.
            init_lr (float): Initial learning rate.
            warmup_steps (int): Number of steps for the warm-up phase.
        """
        self.optimizer = optimizer
        self.total_steps = T_max
        self.max_lr = max_lr
        self.min_lr = min_lr
        self.init_lr = init_lr
        self.warmup_steps = warmup_steps

        self.step_t: int = 0
        self.base_lrs: List[float] = []

        # record current value in self._last_lr to match API from torch.optim.lr_scheduler
        self.last_lr: List[float] = [init_lr]

        self.validate_parameters()

        self._init_lr()

    def validate_parameters(self):
        """
        Validate the parameters to ensure they are non-negative.

        Raises:
            NegativeLRError: If any of the learning rates are negative.
            NegativeStepError: If any of the step values are negative.
        """
        if self.min_lr < 0:
            raise NegativeLRError(self.min_lr, "min_lr")

        if self.max_lr < 0:
            raise NegativeLRError(self.max_lr, "max_lr")

        if self.init_lr < 0:
            raise NegativeLRError(self.init_lr, "init_lr")

        if self.total_steps < 0:
            raise NegativeStepError(self.total_steps, "T_max")

        if self.warmup_steps < 0:
            raise NegativeStepError(self.warmup_steps, "warmup_steps")

    def _init_lr(self):
        """
        Initialize the learning rate for each parameter group in the optimizer.
        """
        self.base_lrs = []
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = self.min_lr
            self.base_lrs.append(self.min_lr)

    def step(self):
        """
        Update the learning rate for the current step.

        Returns:
            float: The updated learning rate.
        """
        if self.step_t < self.warmup_steps:
            value = (
                self.init_lr
                + (self.max_lr - self.init_lr) * self.step_t / self.warmup_steps
            )
        elif self.step_t == self.warmup_steps:
            value = self.max_lr
        else:
            value = self._step()

        self.step_t += 1

        if self.optimizer is not None:
            for param_group in self.optimizer.param_groups:
                param_group["lr"] = value

        self.last_lr = [value]

        return value

    @abstractmethod
    def _step(self) -> float:  # pragma: no cover
        """
        Abstract method to calculate the learning rate for the current step.

        Returns:
            float: The calculated learning rate.
        """
        raise NotImplementedError

    def get_lr(self) -> float:
        """
        Get the current learning rate.

        Returns:
            float: The current learning rate.
        """
        return self.last_lr[0]


class LinearWarmupScheduler(BaseLinearWarmupScheduler):
    r"""Linear LR Scheduler w/ linear warmup."""

    def _step(self) -> float:
        """
        Calculate the learning rate for the current step using a linear decay.

        Returns:
            float: The calculated learning rate.
        """
        return self.max_lr + (self.min_lr - self.max_lr) * (
            self.step_t - self.warmup_steps
        ) / (self.total_steps - self.warmup_steps)


class CosineDecayWithWarmup(BaseLinearWarmupScheduler):
    r"""Cosine LR Scheduler w/ linear warmup."""

    def _step(self) -> float:
        """
        Calculate the learning rate for the current step using a cosine decay.

        Returns:
            float: The calculated learning rate.
        """
        phase: float = (
            (self.step_t - self.warmup_steps)
            / (self.total_steps - self.warmup_steps)
            * math.pi
        )
        return self.min_lr + (self.max_lr - self.min_lr) * (np.cos(phase) + 1.0) / 2.0


class PolySchedulerWithWarmup(BaseLinearWarmupScheduler):
    r"""Poly LR Scheduler.

    Args:
        poly_order (float): LR scheduler decreases with steps.
    """

    def __init__(self, optimizer, poly_order: float = 0.5, **kwargs):
        """
        Initialize the PolySchedulerWithWarmup.

        Args:
            optimizer (torch.optim.Optimizer): Optimizer to apply the learning rate schedule.
            poly_order (float): Order of the polynomial for the learning rate decay.
            kwargs: Additional arguments for the base class.

        Raises:
            ValueError: If poly_order is not positive.
        """
        self.poly_order = poly_order

        if poly_order <= 0:
            raise ValueError(f"[-] poly_order must be positive. {poly_order}")

        super().__init__(optimizer, **kwargs)

    def _step(self) -> float:
        """
        Calculate the learning rate for the current step using a polynomial decay.

        Returns:
            float: The calculated learning rate.
        """
        return (
            self.min_lr
            + (self.max_lr - self.min_lr)
            * (self.step_t - self.warmup_steps) ** self.poly_order
        )



import os
from copy import deepcopy
from typing import TYPE_CHECKING, Optional

from datasets import Dataset, load_dataset, load_from_disk
from lightning.fabric.utilities import rank_zero_only
from tqdm.auto import tqdm

from fusion_bench.utils import timeit_context
import logging

if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer

log = logging.getLogger(__name__)


def load_tokenized_preference_700k_for_rlhf(
    tokenizer: "PreTrainedTokenizer",
    path: str = "hendrydong/preference_700K",
    split: str = "train",
    num_proc: int = 8,
    cache_path: Optional[str] = None,
):
    R"""
    Load and tokenized Preference 700k dataset for Bradley-Terry ranking model.

    The returned dataset contains the following fields:

    - chosen_input_ids: The input token ids for the winner.
    - chosen_attention_mask: The attention mask for the winner.
    - rejected_input_ids: The input token ids for the loser.
    - rejected_attention_mask: The attention mask for the loser.
    """
    if cache_path is not None and os.path.exists(cache_path):
        dataset = load_from_disk(cache_path)
        return dataset

    dataset = load_dataset(path, split=split)

    def tokenize(sample):
        sample["chosen_chat"] = tokenizer.apply_chat_template(
            sample["chosen"], tokenize=False, add_generation_prompt=False
        )
        sample["rejected_chat"] = tokenizer.apply_chat_template(
            sample["rejected"], tokenize=False, add_generation_prompt=False
        )

        tokenized_pos = tokenizer(sample["chosen_chat"], truncation=True)
        tokenized_neg = tokenizer(sample["rejected_chat"], truncation=True)

        # Ensure that the chosen response does not contain an PAD token
        sample["chosen_input_ids"] = tokenized_pos["input_ids"]
        sample["chosen_attention_mask"] = tokenized_pos["attention_mask"]
        if tokenizer.pad_token_id in tokenized_pos["input_ids"]:
            log.warning(f"Prompt contains PAD token: {sample['chosen_chat']}")

        sample["rejected_input_ids"] = tokenized_neg["input_ids"]
        sample["rejected_attention_mask"] = tokenized_neg["attention_mask"]
        # Ensure that the rejected response does not contain an PAD token
        if tokenizer.pad_token_id in tokenized_neg["input_ids"]:
            log.warning(f"Prompt contains PAD token: {sample['rejected_chat']}")

        return sample

    dataset = dataset.map(tokenize, num_proc=num_proc)

    if cache_path is not None and rank_zero_only.rank == 0:
        dataset.save_to_disk(cache_path)
    return dataset



classnames = [
    "AM General Hummer SUV 2000",
    "Acura RL Sedan 2012",
    "Acura TL Sedan 2012",
    "Acura TL Type-S 2008",
    "Acura TSX Sedan 2012",
    "Acura Integra Type R 2001",
    "Acura ZDX Hatchback 2012",
    "Aston Martin V8 Vantage Convertible 2012",
    "Aston Martin V8 Vantage Coupe 2012",
    "Aston Martin Virage Convertible 2012",
    "Aston Martin Virage Coupe 2012",
    "Audi RS 4 Convertible 2008",
    "Audi A5 Coupe 2012",
    "Audi TTS Coupe 2012",
    "Audi R8 Coupe 2012",
    "Audi V8 Sedan 1994",
    "Audi 100 Sedan 1994",
    "Audi 100 Wagon 1994",
    "Audi TT Hatchback 2011",
    "Audi S6 Sedan 2011",
    "Audi S5 Convertible 2012",
    "Audi S5 Coupe 2012",
    "Audi S4 Sedan 2012",
    "Audi S4 Sedan 2007",
    "Audi TT RS Coupe 2012",
    "BMW ActiveHybrid 5 Sedan 2012",
    "BMW 1 Series Convertible 2012",
    "BMW 1 Series Coupe 2012",
    "BMW 3 Series Sedan 2012",
    "BMW 3 Series Wagon 2012",
    "BMW 6 Series Convertible 2007",
    "BMW X5 SUV 2007",
    "BMW X6 SUV 2012",
    "BMW M3 Coupe 2012",
    "BMW M5 Sedan 2010",
    "BMW M6 Convertible 2010",
    "BMW X3 SUV 2012",
    "BMW Z4 Convertible 2012",
    "Bentley Continental Supersports Conv. Convertible 2012",
    "Bentley Arnage Sedan 2009",
    "Bentley Mulsanne Sedan 2011",
    "Bentley Continental GT Coupe 2012",
    "Bentley Continental GT Coupe 2007",
    "Bentley Continental Flying Spur Sedan 2007",
    "Bugatti Veyron 16.4 Convertible 2009",
    "Bugatti Veyron 16.4 Coupe 2009",
    "Buick Regal GS 2012",
    "Buick Rainier SUV 2007",
    "Buick Verano Sedan 2012",
    "Buick Enclave SUV 2012",
    "Cadillac CTS-V Sedan 2012",
    "Cadillac SRX SUV 2012",
    "Cadillac Escalade EXT Crew Cab 2007",
    "Chevrolet Silverado 1500 Hybrid Crew Cab 2012",
    "Chevrolet Corvette Convertible 2012",
    "Chevrolet Corvette ZR1 2012",
    "Chevrolet Corvette Ron Fellows Edition Z06 2007",
    "Chevrolet Traverse SUV 2012",
    "Chevrolet Camaro Convertible 2012",
    "Chevrolet HHR SS 2010",
    "Chevrolet Impala Sedan 2007",
    "Chevrolet Tahoe Hybrid SUV 2012",
    "Chevrolet Sonic Sedan 2012",
    "Chevrolet Express Cargo Van 2007",
    "Chevrolet Avalanche Crew Cab 2012",
    "Chevrolet Cobalt SS 2010",
    "Chevrolet Malibu Hybrid Sedan 2010",
    "Chevrolet TrailBlazer SS 2009",
    "Chevrolet Silverado 2500HD Regular Cab 2012",
    "Chevrolet Silverado 1500 Classic Extended Cab 2007",
    "Chevrolet Express Van 2007",
    "Chevrolet Monte Carlo Coupe 2007",
    "Chevrolet Malibu Sedan 2007",
    "Chevrolet Silverado 1500 Extended Cab 2012",
    "Chevrolet Silverado 1500 Regular Cab 2012",
    "Chrysler Aspen SUV 2009",
    "Chrysler Sebring Convertible 2010",
    "Chrysler Town and Country Minivan 2012",
    "Chrysler 300 SRT-8 2010",
    "Chrysler Crossfire Convertible 2008",
    "Chrysler PT Cruiser Convertible 2008",
    "Daewoo Nubira Wagon 2002",
    "Dodge Caliber Wagon 2012",
    "Dodge Caliber Wagon 2007",
    "Dodge Caravan Minivan 1997",
    "Dodge Ram Pickup 3500 Crew Cab 2010",
    "Dodge Ram Pickup 3500 Quad Cab 2009",
    "Dodge Sprinter Cargo Van 2009",
    "Dodge Journey SUV 2012",
    "Dodge Dakota Crew Cab 2010",
    "Dodge Dakota Club Cab 2007",
    "Dodge Magnum Wagon 2008",
    "Dodge Challenger SRT8 2011",
    "Dodge Durango SUV 2012",
    "Dodge Durango SUV 2007",
    "Dodge Charger Sedan 2012",
    "Dodge Charger SRT-8 2009",
    "Eagle Talon Hatchback 1998",
    "FIAT 500 Abarth 2012",
    "FIAT 500 Convertible 2012",
    "Ferrari FF Coupe 2012",
    "Ferrari California Convertible 2012",
    "Ferrari 458 Italia Convertible 2012",
    "Ferrari 458 Italia Coupe 2012",
    "Fisker Karma Sedan 2012",
    "Ford F-450 Super Duty Crew Cab 2012",
    "Ford Mustang Convertible 2007",
    "Ford Freestar Minivan 2007",
    "Ford Expedition EL SUV 2009",
    "Ford Edge SUV 2012",
    "Ford Ranger SuperCab 2011",
    "Ford GT Coupe 2006",
    "Ford F-150 Regular Cab 2012",
    "Ford F-150 Regular Cab 2007",
    "Ford Focus Sedan 2007",
    "Ford E-Series Wagon Van 2012",
    "Ford Fiesta Sedan 2012",
    "GMC Terrain SUV 2012",
    "GMC Savana Van 2012",
    "GMC Yukon Hybrid SUV 2012",
    "GMC Acadia SUV 2012",
    "GMC Canyon Extended Cab 2012",
    "Geo Metro Convertible 1993",
    "HUMMER H3T Crew Cab 2010",
    "HUMMER H2 SUT Crew Cab 2009",
    "Honda Odyssey Minivan 2012",
    "Honda Odyssey Minivan 2007",
    "Honda Accord Coupe 2012",
    "Honda Accord Sedan 2012",
    "Hyundai Veloster Hatchback 2012",
    "Hyundai Santa Fe SUV 2012",
    "Hyundai Tucson SUV 2012",
    "Hyundai Veracruz SUV 2012",
    "Hyundai Sonata Hybrid Sedan 2012",
    "Hyundai Elantra Sedan 2007",
    "Hyundai Accent Sedan 2012",
    "Hyundai Genesis Sedan 2012",
    "Hyundai Sonata Sedan 2012",
    "Hyundai Elantra Touring Hatchback 2012",
    "Hyundai Azera Sedan 2012",
    "Infiniti G Coupe IPL 2012",
    "Infiniti QX56 SUV 2011",
    "Isuzu Ascender SUV 2008",
    "Jaguar XK XKR 2012",
    "Jeep Patriot SUV 2012",
    "Jeep Wrangler SUV 2012",
    "Jeep Liberty SUV 2012",
    "Jeep Grand Cherokee SUV 2012",
    "Jeep Compass SUV 2012",
    "Lamborghini Reventon Coupe 2008",
    "Lamborghini Aventador Coupe 2012",
    "Lamborghini Gallardo LP 570-4 Superleggera 2012",
    "Lamborghini Diablo Coupe 2001",
    "Land Rover Range Rover SUV 2012",
    "Land Rover LR2 SUV 2012",
    "Lincoln Town Car Sedan 2011",
    "MINI Cooper Roadster Convertible 2012",
    "Maybach Landaulet Convertible 2012",
    "Mazda Tribute SUV 2011",
    "McLaren MP4-12C Coupe 2012",
    "Mercedes-Benz 300-Class Convertible 1993",
    "Mercedes-Benz C-Class Sedan 2012",
    "Mercedes-Benz SL-Class Coupe 2009",
    "Mercedes-Benz E-Class Sedan 2012",
    "Mercedes-Benz S-Class Sedan 2012",
    "Mercedes-Benz Sprinter Van 2012",
    "Mitsubishi Lancer Sedan 2012",
    "Nissan Leaf Hatchback 2012",
    "Nissan NV Passenger Van 2012",
    "Nissan Juke Hatchback 2012",
    "Nissan 240SX Coupe 1998",
    "Plymouth Neon Coupe 1999",
    "Porsche Panamera Sedan 2012",
    "Ram C/V Cargo Van Minivan 2012",
    "Rolls-Royce Phantom Drophead Coupe Convertible 2012",
    "Rolls-Royce Ghost Sedan 2012",
    "Rolls-Royce Phantom Sedan 2012",
    "Scion xD Hatchback 2012",
    "Spyker C8 Convertible 2009",
    "Spyker C8 Coupe 2009",
    "Suzuki Aerio Sedan 2007",
    "Suzuki Kizashi Sedan 2012",
    "Suzuki SX4 Hatchback 2012",
    "Suzuki SX4 Sedan 2012",
    "Tesla Model S Sedan 2012",
    "Toyota Sequoia SUV 2012",
    "Toyota Camry Sedan 2012",
    "Toyota Corolla Sedan 2012",
    "Toyota 4Runner SUV 2012",
    "Volkswagen Golf Hatchback 2012",
    "Volkswagen Golf Hatchback 1991",
    "Volkswagen Beetle Hatchback 2012",
    "Volvo C30 Hatchback 2012",
    "Volvo 240 Sedan 1993",
    "Volvo XC90 SUV 2007",
    "smart fortwo Convertible 2012",
]

templates = [
    lambda c: f"a photo of a {c}.",
    lambda c: f"a photo of the {c}.",
    lambda c: f"a photo of my {c}.",
    lambda c: f"i love my {c}!",
    lambda c: f"a photo of my dirty {c}.",
    lambda c: f"a photo of my clean {c}.",
    lambda c: f"a photo of my new {c}.",
    lambda c: f"a photo of my old {c}.",
]



# flake8: noqa F401
from .clip_fisher_merging import FisherMergingForCLIPVisionModel
from .fisher_merging import FisherMergingAlgorithm, get_param_names_to_merge
from .gpt2_fisher_merging import FisherMergingAlgorithmForGPT2



import hydra.core.hydra_config


def get_hydra_output_dir():
    hydra_cfg = hydra.core.hydra_config.HydraConfig.get()
    return hydra_cfg.runtime.output_dir


def config_priority_get(priority_config, general_config, key, default):
    """
    Retrieve a configuration value with priority.

    This function retrieves the value associated with `key` from `priority_config` if it exists.
    If the key is not found in `priority_config`, it retrieves the value from `general_config`.
    If the key is not found in either configuration, it returns the provided `default` value.

    Args:
        priority_config (dict): The configuration dictionary with higher priority.
        general_config (dict): The general configuration dictionary.
        key (str): The key to look up in the configuration dictionaries.
        default: The default value to return if the key is not found in either configuration.

    Returns:
        The value associated with `key` from `priority_config` or `general_config`, or the `default` value if the key is not found.
    """
    if key in priority_config:
        return priority_config[key]
    return general_config.get(key, default)



import itertools
from typing import Dict, List, Literal, Optional, cast

import lightning as L
import torch.nn.functional as F
from torch import Tensor, nn
from torchmetrics import Metric

from fusion_bench.metrics.nyuv2 import metric_classes
from fusion_bench.metrics.nyuv2.loss import loss_fn


class NYUv2Model(nn.Module):
    image_size = (288, 384)

    def __init__(
        self,
        encoder: nn.Module,
        decoders: nn.ModuleDict,
    ):
        R"""
        Args:
            encoder: The encoder module.
            decoders: A dictionary of the decoder modules.
        """
        super().__init__()

        self.encoder = encoder
        self.decoders = decoders

    def encode(self, images: Tensor) -> Tensor:
        return self.encoder(images)

    def decode(self, features: Tensor, key: str) -> Tensor:
        return self.decoders[key](features)

    def model_parameters(self):
        "parameters of the encoder and the decoders"
        return itertools.chain(
            self.encoder.parameters(),
            self.decoders.parameters(),
        )

    def _compute_loss(
        self,
        outputs: Dict[str, Tensor],
        targets: Dict[str, Tensor],
        task: Literal["segmentation", "depth", "normal"] | List[str],
    ) -> Tensor | List[Tensor]:
        if isinstance(task, str):
            return loss_fn[task](outputs[task], targets[task])
        else:
            return [loss_fn[t](outputs[t], targets[t]) for t in task]

    def forward(
        self, images: Tensor, tasks: Optional[List[str]] = None
    ) -> Dict[str, Tensor]:
        features = self.encode(images)

        if tasks is None:
            tasks = self.decoders.keys()
        outputs = {}
        for task in tasks:
            outputs[task] = F.interpolate(
                self.decode(features, task),
                self.image_size,
                mode="bilinear",
                align_corners=True,
            )

        return outputs


class NYUv2MTLModule(NYUv2Model, L.LightningModule):
    R"""
    multi-task learning module for NYUv2 dataset, with weighted loss
    """

    def __init__(
        self,
        encoder: nn.Module,
        decoders: nn.ModuleDict,
        tasks: List[str],
        task_weights: List[float],
    ):
        R"""
        By using this module, you optimize the weighted sum of the losses of the tasks.
        The joint loss is defined as:

        .. math::
            \mathcal{L} = \sum_{i=1}^{N} w_i \mathcal{L}_i

        where :math:`N` is the number of tasks, :math:`w_i` is the weight for the task :math:`i`,

        Args:
            encoder (nn.Module): The encoder module.
            decoders (nn.ModuleDict): A dictionary of the decoder modules.
            tasks (List[str]): A list of tasks.
            task_weights (List[float]): A list of weights for each task.

        Raises:
            AssertionError: If tasks and task_weights do not have the same length.
        """
        assert len(tasks) == len(
            task_weights
        ), "tasks and task_weights must have the same length"

        super().__init__(encoder, decoders)
        self.tasks = tasks
        self.task_weights = task_weights

        self.metrics = nn.ModuleDict({t: metric_classes[t]() for t in self.tasks})

    # training

    def on_train_epoch_start(self):
        for t in self.tasks:
            self.metrics[t].reset()

    def _single_step(self, images, targets):
        outputs = self(images, self.tasks)
        losses = self._compute_loss(outputs, targets, self.tasks)
        weighted_loss = sum([w * l for w, l in zip(self.task_weights, losses)])

        for t in self.metrics:
            self.metrics[t].update(outputs[t], targets[t])

        return {
            "outputs": outputs,
            "losses": losses,
            "weighted_loss": weighted_loss,
        }

    def training_step(self, batch, batch_idx: int):
        images, targets = batch
        results = self._single_step(images, targets)

        outputs = results["outputs"]
        losses = results["losses"]
        weighted_loss = results["weighted_loss"]

        for i, t in enumerate(self.tasks):
            self.log(f"train/{t}_loss", losses[i])
        self.log("train/loss", weighted_loss, prog_bar=True)

        return weighted_loss

    def on_train_epoch_end(self) -> None:
        for t in self.tasks:
            metrics = cast(Metric, self.metrics[t]).compute()
            for metric_name, metric_value in zip(self.metrics[t].metric_names, metrics):
                self.log(f"train/{t}_{metric_name}", metric_value)

    # validation

    def on_validation_epoch_start(self) -> None:
        for t in self.tasks:
            self.metrics[t].reset()

    def validation_step(self, batch, batch_idx: int):
        images, targets = batch
        results = self._single_step(images, targets)

        outputs = results["outputs"]
        losses = results["losses"]
        weighted_loss = results["weighted_loss"]

        for i, t in enumerate(self.tasks):
            self.log(f"val/{t}_loss", losses[i])
        self.log("val/loss", weighted_loss, prog_bar=True)

    def on_validation_epoch_end(self) -> None:
        for t in self.tasks:
            metrics = cast(Metric, self.metrics[t]).compute()
            for metric_name, metric_value in zip(self.metrics[t].metric_names, metrics):
                self.log(f"val/{t}_{metric_name}", metric_value)



# Copyright (c) 2020, 2021 The HuggingFace Team
# Copyright (c) 2021 Philip May, Deutsche Telekom AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy-Imports module.

This is code taken from the `HuggingFace team <https://huggingface.co/>`__.
Many thanks to HuggingFace for
`your consent <https://github.com/huggingface/transformers/issues/12861#issuecomment-886712209>`__
to publish it as a standalone package.
"""

import importlib
import os
from types import ModuleType
from typing import Any


class LazyImporter(ModuleType):
    """Do lazy imports."""

    # Very heavily inspired by optuna.integration._IntegrationModule
    # https://github.com/optuna/optuna/blob/master/optuna/integration/__init__.py
    def __init__(self, name, module_file, import_structure, extra_objects=None):
        super().__init__(name)
        self._modules = set(import_structure.keys())
        self._class_to_module = {}
        for key, values in import_structure.items():
            for value in values:
                self._class_to_module[value] = key
        # Needed for autocompletion in an IDE
        self.__all__ = list(import_structure.keys()) + sum(
            import_structure.values(), []
        )
        self.__file__ = module_file
        self.__path__ = [os.path.dirname(module_file)]
        self._objects = {} if extra_objects is None else extra_objects
        self._name = name
        self._import_structure = import_structure

    # Needed for autocompletion in an IDE
    def __dir__(self):
        return super().__dir__() + self.__all__

    def __getattr__(self, name: str) -> Any:
        if name in self._objects:
            return self._objects[name]
        if name in self._modules:
            value = self._get_module(name)
        elif name in self._class_to_module:
            module = self._get_module(self._class_to_module[name])
            value = getattr(module, name)
        else:
            raise AttributeError(f"module {self.__name__} has no attribute {name}")

        setattr(self, name, value)
        return value

    def _get_module(self, module_name: str):
        return importlib.import_module("." + module_name, self.__name__)

    def __reduce__(self):
        return (self.__class__, (self._name, self.__file__, self._import_structure))



# NOTE: Working in progress.
import logging
from pathlib import Path
from typing import Any, Literal, Optional, Union  # noqa: F401

import torch
from omegaconf import DictConfig
from PIL.Image import Image
from torch import nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from transformers import (
    AutoFeatureExtractor,
    CLIPProcessor,
    PreTrainedModel,
    ResNetForImageClassification,
)

from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.method import BaseAlgorithm
from fusion_bench.method.adamerging.entropy_loss import entropy_loss
from fusion_bench.mixins import CLIPClassificationMixin
from fusion_bench.modelpool import CLIPVisionModelPool
from fusion_bench.utils import timeit_context
from fusion_bench.utils.data import InfiniteDataLoader
from fusion_bench.utils.instantiate import instantiate

from .warppers.dawe_model import DataAdaptiveWeightEnsemblingCLIPVisionModel

log = logging.getLogger(__name__)


def convert_to_rgb(image: Image | list[Image]) -> Image | list[Image]:
    if isinstance(image, (list, tuple)):
        return [convert_to_rgb(img) for img in image]
    else:
        return image.convert("RGB")


def load_resnet_processor(pretrained_model_name_or_path: str):
    """
    Load a ResNet processor for image preprocessing.

    Args:
        pretrained_model_name_or_path (str): The path or name of the pretrained ResNet model.

    Returns:
        function: A function that processes images using the ResNet processor.
    """
    processor = AutoFeatureExtractor.from_pretrained(pretrained_model_name_or_path)
    return lambda img: processor(
        images=convert_to_rgb(img), return_tensors="pt", do_rescale=False
    ).pixel_values


class ResNetFeatureExtractor(nn.Module):
    def __init__(self, pretrained_model_name_or_path):
        super().__init__()
        self.model = ResNetForImageClassification.from_pretrained(
            pretrained_model_name_or_path
        )
        self.model.classifier = nn.Flatten(1, -1)
        self.config = self.model.config

    def forward(self, *args, **kwargs):
        outputs = self.model(*args, **kwargs)
        return outputs.logits


def load_resnet_feature_extractor(pretrained_model_name_or_path: str):
    model = ResNetFeatureExtractor(pretrained_model_name_or_path)
    return model


def raw_image_collate_fn(batch):
    images, labels = tuple(zip(*batch))
    labels = torch.as_tensor(labels)
    return images, labels


class DataAdaptiveWeightEnsemblingForCLIP(
    BaseAlgorithm,
    CLIPClassificationMixin,
):
    modelpool: CLIPVisionModelPool
    _processor: CLIPProcessor

    def __init__(
        self,
        # merge options
        merge_mode: Literal["task_wise", "layer_wise"],
        init_lambda: float,
        batch_reduce: bool,
        eval_batch_reduce: bool,
        # model options
        dict_processor: DictConfig,
        dict_feature_extractor: DictConfig,
        hidden_size: Optional[int],
        gate_hidden_layers: int,
        task_vector_dtype: Optional[str | torch.dtype],
        task_vector_sparsity: float,
        # training & logging args
        max_steps: int,
        save_interval: int,
        learning_rate: float = 1e-5,
        skip_training: bool = False,
        resume_checkpoint_path: Optional[str] = None,
        # dataloader args
        batch_size: int = 4,
        num_workers: int = 0,
        pin_memory: bool = True,
        **kwargs,
    ):
        # merge options
        self.merge_mode = merge_mode
        self.init_lambda = init_lambda
        self.batch_reduce = batch_reduce
        self.eval_batch_reduce = eval_batch_reduce
        # model options
        self._dict_processor = dict_processor
        self._dict_feature_extractor = dict_feature_extractor
        self.hidden_size = hidden_size
        self.gate_hidden_layers = gate_hidden_layers
        self.task_vector_dtype = task_vector_dtype
        self.task_vector_sparsity = task_vector_sparsity
        # training & logging args
        self.max_steps = max_steps
        self.save_interval = save_interval
        self.learning_rate = learning_rate
        self.skip_training = skip_training
        self.resume_checkpoint_path = resume_checkpoint_path
        # dataloader args
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        super().__init__(**kwargs)

    def load_models(self):
        modelpool = self.modelpool

        dict_processor = instantiate(self._dict_processor)
        clip_processor = modelpool.load_processor()

        dict_feature_extractor: Union[PreTrainedModel, nn.Module] = instantiate(
            self._dict_feature_extractor
        )
        if self.hidden_size is None:
            # try to infer hidden size from feature extractor model
            self.hidden_size = dict_feature_extractor.config.hidden_sizes[-1]

        # initialize classification head
        self.setup_zero_shot_classification_head(
            clip_processor=clip_processor,
            task_names=modelpool.model_names,
        )
        model = DataAdaptiveWeightEnsemblingCLIPVisionModel(
            merge_mode=self.merge_mode,
            hidden_size=self.hidden_size,
            dict_processor=dict_processor,
            model_processor=lambda images: clip_processor(
                images=images, return_tensors="pt"
            ).pixel_values,
            collate_fn=lambda outputs: torch.cat(
                [out.pooler_output for out in outputs], dim=0
            ),
            dict_feature_extractor=dict_feature_extractor,
            base_model=modelpool.load_model("_pretrained_"),
            expert_models=list(modelpool.models()),
            task_vector_dtype=self.task_vector_dtype,
            task_vector_sparsity=self.task_vector_sparsity,
            init_lambda=self.init_lambda,
            gate_hidden_layers=self.gate_hidden_layers,
            batch_reduce=self.batch_reduce,
        )

        if self.resume_checkpoint_path is not None:
            self.fabric.load(self.resume_checkpoint_path, {"model": model})
        return model

    def load_datasets(self):
        modelpool = self.modelpool
        self.test_datasets = {
            task_name: CLIPDataset(
                modelpool.load_test_dataset(task_name),
                processor=None,  # NOTE: processor is not used in CLIPDataset because feature extractor and model may have different processors, so we want to pass the image as is
            )
            for task_name in modelpool.model_names
        }

        # setup dataloaders for test-time adaptation training

        dataloader_kwargs = {
            "batch_size": self.batch_size,
            "num_workers": self.num_workers,
            "pin_memory": self.pin_memory,
        }
        self.shuffled_test_loaders = {
            task_name: self.fabric.setup_dataloaders(
                DataLoader(
                    test_dataset,
                    **dataloader_kwargs,
                    collate_fn=raw_image_collate_fn,
                    shuffle=True,
                )
            )
            for task_name, test_dataset in self.test_datasets.items()
        }
        self.shuffled_test_loader_iters = {
            task_name: InfiniteDataLoader(loader)
            for task_name, loader in self.shuffled_test_loaders.items()
        }

    def run(self, modelpool: CLIPVisionModelPool):
        self.modelpool = modelpool
        with timeit_context("Loading models"):
            model = self.load_models()
        with timeit_context("Loading dataloaders"):
            self.load_datasets()

        # run test-time adaptation
        if not self.skip_training:
            model = self.test_time_adaptation_training(modelpool, model)

        if self.eval_batch_reduce is not None:
            model.batch_reduce = self.eval_batch_reduce
        return model

    def test_time_adaptation_training(self, modelpool, model):
        optimizer = torch.optim.Adam(
            [p for p in model.gate.parameters() if p.requires_grad],
            lr=self.learning_rate,
        )
        model, optimizer = self.fabric.setup(model, optimizer)
        model.train()
        for step_idx in tqdm(
            range(self.max_steps),
            desc="TTA Training",
            dynamic_ncols=True,
        ):
            log_metrics = {}
            losses = 0
            for task_idx, task_name in enumerate(modelpool.model_names):
                # labels are used for logging acc, not involved in training
                images, labels = next(self.shuffled_test_loader_iters[task_name])
                logits = self.compute_logits(model, images=images, task=task_name)
                loss = entropy_loss(logits)
                losses += loss
                log_metrics[f"train/{task_name}_loss"] = loss.item()
                log_metrics[f"train/{task_name}_accuracy"] = (
                    logits.argmax(dim=-1).eq(labels).float().mean().item()
                )

            optimizer.zero_grad()
            self.fabric.backward(losses)
            optimizer.step()

            log_metrics["train/loss"] = losses.item()
            self.fabric.log_dict(log_metrics, step=step_idx)

            if (step_idx + 1) % self.save_interval == 0:
                log.info(f"Saving model at step {step_idx}")
                self.fabric.save(
                    Path(self.log_dir) / "checkpoints" / f"model_{step_idx}.pt",
                    {"model": model},
                )

        if (step_idx + 1) % self.save_interval != 0:
            # if the last step was not saved, save it now
            self.fabric.save(
                Path(self.log_dir) / "checkpoints" / f"model_{step_idx}.pt",
                {"model": model},
            )

        return model



"""
This is the dummy task pool that is used for debugging purposes.
"""

from typing import Optional

from torch import nn

from fusion_bench.models.separate_io import separate_save
from fusion_bench.taskpool.base_pool import BaseTaskPool
from fusion_bench.utils import timeit_context
from fusion_bench.utils.parameters import count_parameters, print_parameters


def get_model_summary(model: nn.Module) -> dict:
    """
    Generate a report for the given model.

    Args:
        model: The model to generate the report for.

    Returns:
        dict: The generated report.
    """
    report = {}
    training_params, all_params = count_parameters(model)
    report["model_info"] = {
        "trainable_params": training_params,
        "all_params": all_params,
        "trainable_percentage": training_params / all_params,
    }
    return report


class DummyTaskPool(BaseTaskPool):
    """
    This is a dummy task pool used for debugging purposes. It inherits from the base TaskPool class.
    """

    def __init__(self, model_save_path: Optional[str] = None):
        super().__init__()
        self.model_save_path = model_save_path

    def evaluate(self, model):
        """
        Evaluate the given model.
        This method does nothing but print the parameters of the model in a human-readable format.

        Args:
            model: The model to evaluate.
        """
        print_parameters(model, is_human_readable=True)

        if self.model_save_path is not None:
            with timeit_context(f"Saving the model to {self.model_save_path}"):
                separate_save(model, self.model_save_path)

        return get_model_summary(model)



import re
from typing import Literal

import datasets
from datasets import load_dataset


def load_gsm8k_question_label_data(
    dataset_name: Literal["train", "test", "train_socratic", "test_socratic"]
):
    R"""
    Load the GSM8K dataset and extract questions and labels.

    An example in the dataset:

    {'question': 'Natalia sold clips to 48 of her friends in April, and then she sold half as many clips in May. How many clips did Natalia sell altogether in April and May?',
     'answer': 'Natalia sold 48/2 = <<48/2=24>>24 clips in May.\nNatalia sold 48+24 = <<48+24=72>>72 clips altogether in April and May.\n#### 72'}

    Args:
        dataset_name (Literal["train", "test", "train_socratic", "test_socratic"]): The name of the dataset to load.

    Returns:
        questions (List[str]): List of questions.
        labels (List[float]): List of labels. For example, the label for the above example is `72.0`.
    """
    # Load the dataset
    if dataset_name in ["train", "test"]:
        dataset = load_dataset("openai/gsm8k", "main", split=dataset_name)
    elif dataset_name in ["train_socratic", "test_socratic"]:
        dataset = load_dataset(
            "openai/gsm8k", "socratic", split=dataset_name.split("_")[0]
        )
    else:
        raise ValueError(f"Unknown dataset name: {dataset_name}")

    # Extract the questions and labels
    questions, labels = [], []
    pattern = r"####\s*(\S+)"
    for i, sample in enumerate(dataset):
        matches = re.findall(pattern, sample["answer"])
        labels.append(float(matches[0].replace(",", "")))
        questions.append(sample["question"])

    return questions, labels


def load_gsm8k_question_label_dataset(
    dataset_name: Literal["train", "test", "train_socratic", "test_socratic"]
):
    """
    Load the GSM8K dataset and return it as a Hugging Face Dataset object.

    Args:
        dataset_name (Literal["train", "test", "train_socratic", "test_socratic"]): The name of the dataset to load.

    Returns:
        datasets.Dataset: The loaded dataset with questions and labels.
    """
    question, labels = load_gsm8k_question_label_data(dataset_name)
    return datasets.Dataset.from_dict({"question": question, "label": labels})



from collections import OrderedDict
from numbers import Number
from typing import Dict, List, Union, cast

import torch
from torch import Tensor

from .parameters import check_parameters_all_equal
from .type import StateDictType


def to_device(
    state_dict: StateDictType,
    device: Union[torch.device, str],
    copy: bool = False,
    inplace: bool = False,
):
    if inplace:
        ret_state_dict = state_dict
    else:
        ret_state_dict = OrderedDict()
    for key in state_dict:
        ret_state_dict[key] = cast(Tensor, state_dict[key]).to(
            device, non_blocking=True, copy=copy
        )
    return ret_state_dict


def state_dicts_check_keys(state_dicts: List[StateDictType]):
    """
    Checks that the state dictionaries have the same keys.

    Args:
        state_dicts (List[Dict[str, Tensor]]): A list of dictionaries containing the state of PyTorch models.

    Raises:
        ValueError: If the state dictionaries have different keys.
    """
    # Get the keys of the first state dictionary in the list
    keys = set(state_dicts[0].keys())
    # Check that all the state dictionaries have the same keys
    for state_dict in state_dicts:
        assert keys == set(state_dict.keys()), "keys of state_dicts are not equal"


def num_params_of_state_dict(state_dict: StateDictType):
    """
    Returns the number of parameters in a state dict.

    Args:
        state_dict (Dict[str, Tensor]): The state dict to count the number of parameters in.

    Returns:
        int: The number of parameters in the state dict.
    """
    return sum([state_dict[key].numel() for key in state_dict])


def state_dict_flatten(state_dict: Dict[str, Tensor]):
    """
    Flattens a state dict.

    Args:
        state_dict (Dict[str, Tensor]): The state dict to be flattened.

    Returns:
        Tensor: The flattened state dict.
    """
    flattened_state_dict = []
    for key in state_dict:
        flattened_state_dict.append(state_dict[key].flatten())
    return torch.cat(flattened_state_dict)


def state_dict_avg(state_dicts: List[StateDictType]):
    """
    Returns the average of a list of state dicts.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to average.

    Returns:
        Dict: The average of the state dicts.
    """
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    num_state_dicts = len(state_dicts)
    avg_state_dict = OrderedDict()
    for key in state_dicts[0]:
        avg_state_dict[key] = torch.zeros_like(state_dicts[0][key])
        for state_dict in state_dicts:
            avg_state_dict[key] += state_dict[key]
        avg_state_dict[key] /= num_state_dicts
    return avg_state_dict


def state_dict_sub(
    a: StateDictType, b: StateDictType, strict: bool = True, device=None
):
    """
    Returns the difference between two state dicts `a-b`.

    Args:
        a (StateDictType): The first state dict.
        b (StateDictType): The second state dict.
        strict (bool): Whether to check if the keys of the two state dicts are the same.

    Returns:
        StateDictType: The difference between the two state dicts.
    """
    if strict:
        assert set(a.keys()) == set(b.keys())

    diff = OrderedDict()
    for k in a:
        if k in b:
            diff[k] = a[k] - b[k]
            if device is not None:
                diff[k] = diff[k].to(device, non_blocking=True)
    return diff


def state_dict_add(
    a: StateDictType, b: StateDictType, strict: bool = True, device=None
):
    """
    Returns the sum of two state dicts.

    Args:
        a (Dict): The first state dict.
        b (Dict): The second state dict.
        strict (bool): Whether to check if the keys of the two state dicts are the same.

    Returns:
        Dict: The sum of the two state dicts.
    """
    ans = {}
    if strict:
        check_parameters_all_equal([a, b])
        for key in a:
            ans[key] = a[key] + b[key]
    else:
        for key in a:
            if key in b:
                ans[key] = a[key] + b[key]
    if device is not None:
        ans = to_device(ans, device)
    return ans


def state_dict_add_scalar(a: StateDictType, scalar: Number):
    ans = OrderedDict()
    for key in a:
        ans[key] = a[key] + scalar
    return ans


def state_dict_mul(state_dict: StateDictType, scalar: float):
    """
    Returns the product of a state dict and a scalar.

    Args:
        state_dict (Dict): The state dict to be multiplied.
        scalar (float): The scalar to multiply the state dict with.

    Returns:
        Dict: The product of the state dict and the scalar.
    """
    diff = OrderedDict()
    for k in state_dict:
        diff[k] = scalar * state_dict[k]
    return diff


def state_dict_div(state_dict: StateDictType, scalar: float):
    """
    Returns the division of a state dict by a scalar.

    Args:
        state_dict (Dict): The state dict to be divided.
        scalar (float): The scalar to divide the state dict by.

    Returns:
        Dict: The division of the state dict by the scalar.
    """
    diff = OrderedDict()
    for k in state_dict:
        diff[k] = state_dict[k] / scalar
    return diff


def state_dict_power(state_dict: Dict[str, Tensor], p: float):
    """
    Returns the power of a state dict.

    Args:
        state_dict (Dict[str, Tensor]): The state dict to be powered.
        p (float): The power to raise the state dict to.

    Returns:
        Dict[str, Tensor]: The powered state dict.
    """
    powered_state_dict = {}
    for key in state_dict:
        powered_state_dict[key] = state_dict[key] ** p
    return powered_state_dict


def state_dict_interpolation(
    state_dicts: List[Dict[str, Tensor]], scalars: List[float]
):
    """
    Interpolates between a list of state dicts using a list of scalars.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to interpolate between.
        scalars (List[float]): The list of scalars to use for interpolation.

    Returns:
        Dict: The interpolated state dict.
    """
    assert len(state_dicts) == len(
        scalars
    ), "The number of state_dicts and scalars must be the same"
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    interpolated_state_dict = {}
    for key in state_dicts[0]:
        interpolated_state_dict[key] = torch.zeros_like(state_dicts[0][key])
        for state_dict, scalar in zip(state_dicts, scalars):
            interpolated_state_dict[key] += scalar * state_dict[key]
    return interpolated_state_dict


def state_dict_sum(state_dicts: List[StateDictType]):
    """
    Returns the sum of a list of state dicts.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to sum.

    Returns:
        Dict: The sum of the state dicts.
    """
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    sum_state_dict = OrderedDict()
    for key in state_dicts[0]:
        sum_state_dict[key] = 0
        for state_dict in state_dicts:
            sum_state_dict[key] = sum_state_dict[key] + state_dict[key]
    return sum_state_dict


def state_dict_weighted_sum(
    state_dicts: List[Dict[str, Tensor]], weights: List[float], device=None
):
    """
    Returns the weighted sum of a list of state dicts.

    Args:
        state_dicts (List[Dict[str, Tensor]]): The list of state dicts to interpolate between.
        weights (List[float]): The list of weights to use for the weighted sum.

    Returns:
        Dict: The weighted sum of the state dicts.
    """
    assert len(state_dicts) == len(
        weights
    ), "The number of state_dicts and weights must be the same"
    assert len(state_dicts) > 0, "The number of state_dicts must be greater than 0"
    assert all(
        [len(state_dicts[0]) == len(state_dict) for state_dict in state_dicts]
    ), "All state_dicts must have the same number of keys"

    weighted_sum_state_dict: Dict[str, Tensor] = {}
    for key in state_dicts[0]:
        # states dicts can be sparse matrices
        weighted_sum_state_dict[key] = torch.zeros_like(state_dicts[0][key]).to_dense()
        for state_dict, weight in zip(state_dicts, weights):
            weighted_sum_state_dict[key] = torch.add(
                weighted_sum_state_dict[key], weight * state_dict[key]
            )
        if device is not None:
            weighted_sum_state_dict[key] = weighted_sum_state_dict[key].to(
                device, non_blocking=True
            )
    return weighted_sum_state_dict



import logging
from abc import abstractmethod
from typing import Dict, List, Literal, Optional, Tuple, cast  # noqa: F401

import torch
import torch.utils.hooks
from torch import Tensor, nn
from tqdm.auto import tqdm
from transformers import LlamaForCausalLM

from fusion_bench.method import BaseAlgorithm
from fusion_bench.method.pruning.wanda_utils.data import get_loaders
from fusion_bench.method.pruning.wanda_utils.prune import prepare_calibration_input
from fusion_bench.mixins import SimpleProfilerMixin
from fusion_bench.modelpool import CausalLMPool
from fusion_bench.utils import timeit_context
from fusion_bench.utils.cache_utils import cache_to_disk

from .prune_utils import (
    PruningType,
    compute_sparsity,
    find_linear_layers,
    semistructured_magnitude_prune_,
    unstructured_magnitude_prune_,
)

log = logging.getLogger(__name__)


class BaseLoSparseHookFn:
    """
    Base class for low-sparsity hook functions.
    """

    def __init__(self, linear):
        self.linear = linear

    @abstractmethod
    def compute(self) -> Tensor:
        """
        Compute the importance scores.
        """
        pass

    @abstractmethod
    def __call__(self, linear, inp: Tuple[Tensor], out: Tensor):
        """
        Hook function to be called during the forward pass.
        """
        pass


class WandaHookFn(BaseLoSparseHookFn):
    R"""
    Here in this class, the `scalar_row` is the mean of the squared sum of the input to the linear layer along a specific input dimension.

    $$\frac{\sum_{i=1}^{N L} X_{ij}^2}{N L}$$
    """

    def __init__(self, linear: nn.Linear):
        super().__init__(linear)

        self.scalar_row = torch.zeros(
            (linear.weight.size(1),), device=linear.weight.device
        )
        self.nsamples = 0

    def compute(self):
        return torch.abs(self.linear.weight) * torch.sqrt(
            self.scalar_row.reshape(1, -1)
        )

    def __call__(self, linear: nn.Linear, inps: Tuple[Tensor], out: Tensor):
        assert len(inps) == 1
        inp = inps[0]
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)

        batch_size = inp.shape[0]
        if len(inp.shape) == 3:
            inp = inp.reshape((-1, inp.shape[-1]))
        # (NxL, C) -> (C, NxL)
        inp = inp.t()

        self.scalar_row *= self.nsamples / (self.nsamples + batch_size)
        self.nsamples += batch_size

        inp = inp.type(torch.float32)
        self.scalar_row += torch.norm(inp, p=2, dim=1) ** 2 / self.nsamples


class WandaPruningForLlama(BaseAlgorithm, SimpleProfilerMixin):
    """
    Class for Wanda pruning for Llama models.
    """

    _config_mapping = BaseAlgorithm._config_mapping | {
        "nsamples": "nsamples",
        "seed": "seed",
        "use_variant": "use_variant",
        "prune_type": "prune_type",
        "device": "device",
        "dtype": "dtype",
        "sparsity_ratio": "sparsity_ratio",
        "n": "n",
        "m": "m",
    }

    def __init__(
        self,
        *,
        nsamples: int,
        seed: int,
        use_variant: bool,
        prune_type: PruningType,
        device: str,
        dtype: str,
        sparsity_ratio: float,
        n: int,
        m: int,
        model_save_path: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize the WandaPruningForLlama class.

        Args:
            nsamples (int): Number of samples for calibration.
            seed (int): Random seed.
            use_variant (bool): Whether to use a variant of the pruning method.
            prune_type (PruningType): Type of pruning to perform.
            device (str): Device to use for computation.
            dtype (str): Data type to use for computation.
            sparsity_ratio (float): Sparsity ratio for pruning.
            n (int): Number of elements to keep in semi-structured pruning.
            m (int): Number of elements in a group for semi-structured pruning.
            model_save_path (Optional[str]): Path to save the pruned model.
            **kwargs: Additional arguments.
        """
        super().__init__(**kwargs)
        self.nsamples = nsamples
        self.seed = seed
        self.use_variant = use_variant
        self.prune_type = prune_type
        self.device = device
        self.dtype = dtype
        self.sparsity_ratio = sparsity_ratio
        self.n = n
        self.m = m
        self.model_save_path = model_save_path

    def run(self, modelpool: CausalLMPool):
        """
        Run the pruning algorithm on the model pool.

        Args:
            modelpool (CausalLMPool): Pool of causal language models.

        Returns:
            LlamaForCausalLM: Pruned model.
        """

        # load pre-trained model or the first model in the pool
        with self.profile("load_model"):
            model = modelpool.load_pretrained_or_first_model()
            model.seqlen = model.config.max_position_embeddings
            tokenizer = modelpool.load_tokenizer(use_fast=False)

        if not isinstance(model, (LlamaForCausalLM,)):
            log.warning(f"Model type {type(model)} may not supported.")

        inps, outs, attention_mask, position_ids = self.prepare_calibration_data(
            model, tokenizer
        )

        self.prune_using_calibration_data_(
            model,
            inps=inps,
            outs=outs,
            attention_mask=attention_mask,
            position_ids=position_ids,
        )

        if self.model_save_path is not None:
            with timeit_context(f"Saving pruned model to {self.model_save_path}"):
                tokenizer.save_pretrained(self.model_save_path)
                model.save_pretrained(self.model_save_path)
        return model

    def _prepare_calibration_data(self, model, tokenizer):
        """
        Prepare calibration data for pruning.

        Args:
            model (LlamaForCausalLM): Model to be pruned.
            tokenizer: Tokenizer for the model.

        Returns:
            Tuple: Calibration data (inputs, outputs, attention mask, position IDs).
        """
        with timeit_context("loading calibration data"):
            dataloader, _ = get_loaders(
                "c4",
                nsamples=self.nsamples,
                seed=self.seed,
                seqlen=model.seqlen,
                tokenizer=tokenizer,
            )

        with torch.no_grad():
            # collect input to the first layer
            inps, outs, attention_mask, position_ids = prepare_calibration_input(
                model, dataloader, self.device
            )
        return inps, outs, attention_mask, position_ids

    def prepare_calibration_data(self, model: LlamaForCausalLM, tokenizer):
        """
        Prepare calibration data for pruning with caching.

        Args:
            model (LlamaForCausalLM): Model to be pruned.
            tokenizer: Tokenizer for the model.

        Returns:
            Tuple: Calibration data (inputs, outputs, attention mask, position IDs).
        """

        @cache_to_disk(
            f"outputs/cache/{model.config.name_or_path.split('/')[-1]}/calibration_data.pkl"
        )
        def _prepare_calibration_data(model, tokenizer):
            return self._prepare_calibration_data(model, tokenizer)

        return _prepare_calibration_data(model, tokenizer)

    def prune_using_calibration_data_(
        self,
        model: LlamaForCausalLM,
        *,
        inps,
        outs,
        attention_mask,
        position_ids,
    ):
        """
        Prune the model using calibration data.

        Args:
            model (LlamaForCausalLM): Model to be pruned.
            inps: Calibration inputs.
            outs: Calibration outputs.
            attention_mask: Attention mask for calibration data.
            position_ids: Position IDs for calibration data.
        """
        layers = model.model.layers
        for layer_idx, layer in tqdm(
            enumerate(layers),
            "Pruning Layers",
            total=len(layers),
            dynamic_ncols=True,
        ):
            if (
                hasattr(model, "hf_device_map")
                and f"model.layers.{layer_idx}" in model.hf_device_map
            ):
                # handle the case for llama-30B and llama-65B, when the device map has multiple GPUs;
                dev = model.hf_device_map[f"model.layers.{layer_idx}"]
                inps, outs, attention_mask, position_ids = (
                    inps.to(dev),
                    outs.to(dev),
                    attention_mask.to(dev) if attention_mask is not None else None,
                    position_ids.to(dev) if position_ids is not None else None,
                )

            # collect the importance scores
            linear_layers = cast(
                Dict[str, nn.Linear],
                find_linear_layers(layer, layers=[nn.Linear]),
            )

            # register hooks to collect the importance scores
            def get_hook_fn(linear: nn.Linear):
                hook_fn = WandaHookFn(linear)
                return hook_fn

            hooks = {}
            handles: List[torch.utils.hooks.RemovableHandle] = []
            for name, linear in linear_layers.items():
                hook_fn = get_hook_fn(linear)
                hooks[name] = hook_fn
                handles.append(linear.register_forward_hook(hook_fn))

            with torch.no_grad():
                for j in range(self.nsamples):
                    outs[j] = layer(
                        inps[j].unsqueeze(0),
                        attention_mask=attention_mask,
                        position_ids=position_ids,
                    )[0]

            # compute the importance scores and remove the hooks
            metrics = {}
            for name, hook in hooks.items():
                metrics[name] = hook.compute()
            for h in handles:
                h.remove()

            # prune the weights based on the importance scores
            if self.prune_type == PruningType.UNSTRUCTURED:
                for name, linear in linear_layers.items():
                    log.info(f"Pruning {name}")
                    unstructured_magnitude_prune_(
                        linear.weight.data,
                        metrics[name],
                        sparsity_ratio=self.sparsity_ratio,
                    )
                    self.check_sparsity(linear.weight)
            elif self.prune_type == PruningType.SEMISTRUCTURED:
                for name, linear in linear_layers.items():
                    log.info(f"Pruning {name}")
                    semistructured_magnitude_prune_(
                        linear.weight.data,
                        metrics[name],
                        n=self.n,
                        m=self.m,
                    )
                    self.check_sparsity(linear.weight)
            else:
                raise ValueError(f"Invalid pruning type: {self.prune_type}")

            # compute the input to the next layer
            with torch.no_grad():
                for j in range(self.nsamples):
                    outs[j] = layer(
                        inps[j].unsqueeze(0),
                        attention_mask=attention_mask,
                        position_ids=position_ids,
                    )[0]
            inps, outs = outs, inps

    @torch.no_grad()
    def check_sparsity(self, weight: Tensor, tol: float = 0.01):
        """
        Check the sparsity of the weight tensor.

        Args:
            weight (Tensor): Weight tensor.
            tol (float): Tolerance for sparsity check.

        Raises:
            ValueError: If the pruning type is invalid.
        """
        if self.prune_type == PruningType.UNSTRUCTURED:
            assert (compute_sparsity(weight) - self.sparsity_ratio).abs() < tol
        elif self.prune_type == PruningType.SEMISTRUCTURED:
            assert (compute_sparsity(weight) - self.n / self.m).abs() < tol
        else:
            raise ValueError(f"Invalid pruning type: {self.prune_type}")



import torch
import torch.nn as nn


# Define WrappedGPT class
class WrappedGPT:
    """
    This class wraps a GPT layer for specific operations.

    Attributes:
        layer (nn.Linear | nn.Module): The GPT layer to be wrapped.
        dev (torch.device): The device on which the layer's weights are stored.
        rows (int): The number of rows in the layer's weight matrix.
        columns (int): The number of columns in the layer's weight matrix.
        scaler_row (torch.Tensor): A tensor to store the scaler values for each column.
        nsamples (int): The number of samples processed.
        layer_id (int): The ID of the layer.
        layer_name (str): The name of the layer.
    """

    def __init__(self, layer: nn.Linear | nn.Module, layer_id=0, layer_name="none"):
        """
        Initialize the WrappedGPT class.

        Args:
            layer (nn.Linear | nn.Module): The GPT layer to be wrapped.
            layer_id (int, optional): The ID of the layer. Defaults to 0.
            layer_name (str, optional): The name of the layer. Defaults to "none".
        """
        self.layer = layer
        self.dev = self.layer.weight.device
        self.rows = layer.weight.data.shape[0]
        self.columns = layer.weight.data.shape[1]

        self.scaler_row = torch.zeros((self.columns), device=self.dev)
        self.nsamples = 0

        self.layer_id = layer_id
        self.layer_name = layer_name

    def add_batch(self, inp: torch.Tensor, out: torch.Tensor):
        """
        Add a batch of input and output tensors to the scaler_row.

        Args:
            inp (torch.Tensor): The input tensor.
            out (torch.Tensor): The output tensor.
        """
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if isinstance(self.layer, nn.Linear):
            if len(inp.shape) == 3:
                inp = inp.reshape((-1, inp.shape[-1]))
            inp = inp.t()

        self.scaler_row *= self.nsamples / (self.nsamples + tmp)
        self.nsamples += tmp

        inp = inp.type(torch.float32)
        self.scaler_row += torch.norm(inp, p=2, dim=1) ** 2 / self.nsamples



import logging
from typing import Dict, List, cast  # noqa: F401

import torch
import torch.utils.data
from omegaconf import DictConfig
from torch import Tensor, nn
from torch.nn.modules import Module
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm

from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.mixins import CLIPClassificationMixin

from .regmean import RegMeanAlgorithm

log = logging.getLogger(__name__)


class RegMeanAlgorithmForCLIP(
    RegMeanAlgorithm,
    CLIPClassificationMixin,
):
    _config_mapping = {
        "_dataloader_kwargs": "dataloader_kwargs",
    }

    def __init__(self, *, dataloader_kwargs: DictConfig, **kwargs):
        super().__init__(**kwargs)
        self._dataloader_kwargs = dataloader_kwargs

    def on_regmean_start(self):
        self.setup_zero_shot_classification_head()

    def compute_logits(self, module, batch, task: str) -> Tensor:
        images, _ = batch
        text_embeds = self.zeroshot_weights[task]

        image_embeds = module(images)[1]
        image_embeds = self.visual_projection(image_embeds)

        # normalize embeddings
        image_embeds = image_embeds / image_embeds.norm(p=2, dim=-1, keepdim=True)

        # cosine similarity
        logits_per_text = (
            torch.matmul(text_embeds, image_embeds.t()) * self.logit_scale_exp
        )
        logits_per_image = logits_per_text.t()

        return logits_per_image

    def get_regmean_weights(
        self,
        model_name: str,
        model: Module,
        train_dataset: torch.utils.data.Dataset,
        linear_modules_to_merge: Dict[str, Module],
    ):
        # setup dataloader
        train_dataset = CLIPDataset(train_dataset, self.clip_processor)
        train_dataloader = DataLoader(
            train_dataset, shuffle=True, **self._dataloader_kwargs
        )
        train_dataloader = self.fabric.setup_dataloaders(train_dataloader)
        model = self.fabric.setup(model)

        def compute_regmean_weights(module_name: str):
            """
            compute the regmean weights, a hook function to deal with each module's input
            :param module_name: str, module name
            :return:
            """

            def hook(module: nn.Module, input: tuple, output: torch.Tensor):
                # Tensor, shape (batch_size, sequence_length, hidden_dim)
                x = cast(Tensor, input[0]).detach()
                batch_num_actual_examples = x.shape[0]
                # Tensor, shape (batch_size * sequence_length, hidden_dim)
                x = x.reshape(-1, x.shape[-1])
                # Tensor, shape (hidden_dim, hidden_dim)
                xtx = torch.matmul(x.transpose(0, 1), x)
                # store the averaged weights in regmean_weights
                if module_name not in regmean_weights.keys():
                    regmean_weights[module_name] = xtx / x.shape[0]
                    num_computed_examples[module_name] = x.shape[0]
                    num_actual_examples[module_name] = batch_num_actual_examples
                else:
                    regmean_weights[module_name] = (
                        regmean_weights[module_name]
                        * num_computed_examples[module_name]
                        + xtx
                    ) / (num_computed_examples[module_name] + x.shape[0])
                    num_computed_examples[module_name] += x.shape[0]
                    num_actual_examples[module_name] += batch_num_actual_examples

            return hook

        handles = []
        # dictionary, regmean matrices for each linear module inputs
        regmean_weights = {}
        # dictionary, number of examples (multiplied the sequence length) used for computing regmean matrices
        num_computed_examples = {}
        # dictionary, number of actual examples used for computing regmean matrices
        num_actual_examples = {}

        for module_name, linear_module_to_merge in linear_modules_to_merge.items():
            # register a hook in the forward process
            handle = linear_module_to_merge.register_forward_hook(
                compute_regmean_weights(module_name=module_name)
            )
            handles.append(handle)
        for step, batch in tqdm(
            enumerate(train_dataloader),
            desc=f"computing regmean weights for model {model_name}",
        ):
            if (
                len(num_actual_examples) > 0
                and list(num_actual_examples.values())[0] >= self.num_regmean_examples
            ):
                break
            logits = self.compute_logits(model, batch, model_name)  # noqa: F841

        # remove the added hook
        for handle in handles:
            handle.remove()

        for module_name in regmean_weights.keys():
            regmean_weights[module_name] = regmean_weights[module_name].detach().cpu()

        return regmean_weights



import logging

from transformers import AutoTokenizer

from .llama_3_Instruct import CHAT_TEMPLATE as LLAMA_3_INSTRUCT_CHAT_TEMPLATE

chat_template_mapping = {"llama_3_instruct": LLAMA_3_INSTRUCT_CHAT_TEMPLATE}

log = logging.getLogger(__name__)


def load_tokenizer_with_chat_template(
    pretrained_model_name_or_path: str,
    model_family: str,
    overwrite_chat_template: bool = True,
    **kwargs,
):
    """
    Load the tokenizer for Llama 3 model.

    Args:
        pretrained_model_name_or_path (str): The name or path of the pretrained model.
        model_family (str): The model family.
        **kwargs: Additional keyword arguments passed to the tokenizer class.
    """
    assert (
        model_family in chat_template_mapping
    ), f"Model family {model_family} not found. Available model families: {chat_template_mapping.keys()}"

    tokenizer = AutoTokenizer.from_pretrained(
        pretrained_model_name_or_path,
        **kwargs,
    )

    if tokenizer.chat_template is None:
        tokenizer.chat_template = chat_template_mapping[model_family]
    else:
        if overwrite_chat_template:
            log.warning("Overwriting the chat template with the default chat template.")
            tokenizer.chat_template = chat_template_mapping[model_family]
        else:
            log.warning("Chat template already exists. Skipping overwriting.")
    return tokenizer



from typing import List, Mapping

import torch
from torch import nn

__all__ = "ParamterDictModel"


def set_attr(obj, names: List[str], val, check_parent: bool = False):
    """
    Sets an attribute of an object recursively.

    Args:
        obj (object): Object to set attribute of.
        names (list): List of attribute names to set recursively.
        val (object): Value to set the attribute to.
        check_parent (bool): If True, checks if the parent attribute exists; otherwise, creates it if it does not exist.
    """
    if len(names) == 1:
        setattr(obj, names[0], val)
    else:
        if check_parent and not hasattr(obj, names[0]):
            setattr(obj, names[0], nn.Module())
        set_attr(getattr(obj, names[0]), names[1:], val, check_parent=check_parent)


def has_attr(obj, names: List[str]):
    """
    Checks if an attribute exists in an object recursively.

    Args:
        obj (object): Object to check attribute of.
        names (list): List of attribute names to check recursively.

    Returns:
        bool: True if the attribute exists; otherwise, False.
    """
    if len(names) == 1:
        return hasattr(obj, names[0])
    else:
        return has_attr(getattr(obj, names[0]), names[1:])


class ParameterDictModel(nn.Module):
    """
    This model is used to create a model with parameters from a dictionary.
    It behaves like a normal `nn.ParameterDict`, but support keys with dots.
    """

    def __init__(
        self,
        parameters: Mapping[str, nn.Parameter],
    ):
        super().__init__()
        for name, param in parameters.items():
            assert isinstance(param, nn.Parameter), f"{name} is not a nn.Parameter"
            set_attr(
                self,
                name.split("."),
                param,
                check_parent=True,
            )

    def __repr__(self):
        """
        Generate a string representation of the model's parameters.

        Returns:
            str: A string representation of the model's parameters.
        """
        param_reprs = []
        for name, param in self.named_parameters():
            param_repr = f"{name}: {param.size()}"
            param_reprs.append(param_repr)
        return f"{self.__class__.__name__}({', '.join(param_reprs)})"



"""
This module contains classes for representing ARC tasks, examples, and grids in different formats.
"""

import re
from abc import ABC, abstractmethod
from io import BytesIO
from types import WrapperDescriptorType
from typing import List, Optional, Text, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import ListedColormap, Normalize
from scipy.ndimage import generate_binary_structure, label

from .arc import Example, Grid, Task
from .np_cache import np_lru_cache

# =============== CONSTANTS ===============
COLUMN_SEP = " "
ROW_SEP = "\n"
IO_SEP = "\n\n"
EXAMPLE_SEP = "\n\n"
TRAIN_HEADER = "==TRAIN==\n"
TRAIN_TEST_SEP = "\n\n"
TEST_HEADER = "==TEST==\n"

# =============== UTILS ===============


def parse_numpy_from_str(array_str: str) -> np.ndarray:
    """
    Parses a string representation of a 2D array into a NumPy ndarray.

    Parameters:
    - array_str (str): A string representation of a 2D array, where rows are separated by newlines.

    Returns:
    - np.ndarray: A NumPy array of type int8 representing the parsed 2D array.
    """
    try:
        # Remove the surrounding brackets from the string
        clean_str = array_str.replace("[", "").replace("]", "")

        # Split the cleaned string by whitespace to get individual elements and convert them to integers
        elements = list(map(int, clean_str.split()))

        # Determine the number of rows by counting the newline characters and adding one
        rows = array_str.count("\n") + 1

        # Calculate the number of columns by dividing the total number of elements by the number of rows
        cols = len(elements) // rows

        # Create the NumPy array with the determined shape and convert it to type int8
        array = np.array(elements).reshape((rows, cols)).astype(np.int8)

        return array
    except Exception as e:
        # Print the exception message and the original string for debugging purposes
        print(e)
        print(array_str)
        # Return a default 1x1 array with a zero element in case of an error
        # raise e
        return None


# =============== INTERFACE ===============


class GridRepresenter(ABC):
    @abstractmethod
    def encode(self, grid: Grid) -> str:
        pass

    @abstractmethod
    def decode(self, encoded_str: str, **kwargs) -> Grid:
        pass

    def display(self, encoded_str: str):
        print(self.decode(encoded_str))


class ExampleRepresenter(ABC):
    grid_representer: GridRepresenter

    @abstractmethod
    def encode(self, example: Example, **kwargs) -> Union[str, Tuple[str, str]]:
        pass

    @abstractmethod
    def decode(self, encoded: Tuple[str, str], **kwargs) -> Example:
        pass

    def display(self, encoded: Union[str, Tuple[str, str]]):
        if isinstance(encoded, str):
            print(encoded)
        else:
            print("\n".join(encoded))


class TaskRepresenter(ABC):
    example_representer: ExampleRepresenter

    @abstractmethod
    def encode(self, task: Task, **kwargs) -> Union[Tuple[str, str, str], str]:
        pass

    @abstractmethod
    def decode(self, encoded: Tuple[str, str], **kwargs) -> Task:
        pass

    def display(self, encoded: Union[str, Tuple[str, str]]):
        if isinstance(encoded, str):
            print(encoded)
        else:
            print("\n".join(encoded))


# =============== GRID REPRESENTATION ===============
class DelimitedGridRepresenter(GridRepresenter):
    def __init__(self, column_sep: str = " ", row_sep: str = "\n"):
        self.column_sep: str = column_sep
        self.row_sep: str = row_sep

    def encode(self, grid: Grid) -> str:
        output = ""
        for i in range(grid.shape[0]):
            for j in range(grid.shape[1]):
                output += str(grid[i][j]) + self.column_sep
            output = output[:-1] + self.row_sep
        return output[: -len(self.row_sep)]

    def decode(self, encoded_str: str) -> Grid:
        rows = encoded_str.strip().split(self.row_sep)
        grid = [list(map(int, row.split(self.column_sep))) for row in rows]
        return np.array(grid)

    def __repr__(self) -> str:
        return f"DelimitedGridRepresenter(column_sep={self.column_sep!r}, row_sep={self.row_sep!r})"

    def __str__(self) -> str:
        return repr(self)


class PythonListGridRepresenter(GridRepresenter):
    def encode(self, grid: Grid) -> str:
        return str(grid)

    def decode(self, encoded_str: str) -> Grid:
        return parse_numpy_from_str(encoded_str)

    def __repr__(self) -> str:
        return "PythonListGridRepresenter()"

    def __str__(self) -> str:
        return repr(self)


# Used in BARC
class WordGridRepresenter(GridRepresenter):
    color_map = {
        0: "Black",
        1: "Blue",
        2: "Red",
        3: "Green",
        4: "Yellow",
        5: "Gray",
        6: "Pink",
        7: "Orange",
        8: "Purple",
        9: "Brown",
    }

    def __init__(self):
        self.inv_map = {v: k for k, v in self.color_map.items()}

    def encode(self, grid: Grid) -> str:
        output = ""
        for i in range(grid.shape[0]):
            for j in range(grid.shape[1]):
                output += self.color_map[grid[i][j]] + " "
            output = output[:-1] + "\n"
        return output[:-1]

    def decode(self, encoded_str: str) -> Grid:
        rows = encoded_str.strip().split("\n")
        grid = [[self.inv_map[color] for color in row.split()] for row in rows]
        return np.array(grid)

    def __str__(self) -> str:
        return "WordGridRepresenter()"

    def __repr__(self) -> str:
        return "WordGridRepresenter()"


# This is adapted from Greenblat 2024
class ConnectedComponentRepresenter(GridRepresenter):
    normalized: bool = True
    max_token_per_color: Optional[int] = None
    disable_absolute: bool = False
    dotsafter: int = 4
    connected_component: int = 4
    sort_by_count: bool = False
    spreadsheet_col_labels: List[str] = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
        "AA",
        "AB",
        "AC",
        "AD",
        "AE",
        "AF",
    ]

    def __init__(
        self,
        normalized: bool = True,
        max_token_per_color: Optional[int] = None,
        disable_absolute: bool = False,
        sort_by_count: bool = False,
    ):
        self.normalized = normalized
        self.max_token_per_color = max_token_per_color
        self.disable_absolute = disable_absolute
        self.sort_by_count = sort_by_count

    def to_spreadsheet(self, i: int, j: int) -> str:
        try:
            out = f"{self.spreadsheet_col_labels[j]}{i+1}"
        except IndexError:
            print(i, j)
            raise
        return out

    def to_spreadsheet_with_dots(self, rows_cols: List[Tuple[int, int]]) -> str:
        row_cols_v = np.array(sorted(rows_cols, key=lambda x: (x[0], x[1])))
        running_str = ""
        idx = 0
        while idx < len(row_cols_v):
            r, c = row_cols_v[idx]
            count_in_a_row = 0
            for checking_idx, (n_r, n_c) in enumerate(row_cols_v[idx:]):
                if n_r == r and n_c == c + checking_idx:
                    count_in_a_row += 1
                else:
                    break
            if count_in_a_row > self.dotsafter:
                start = self.to_spreadsheet(r, c)
                c_end = c + count_in_a_row - 1
                assert np.array_equal(
                    row_cols_v[idx + count_in_a_row - 1], (r, c_end)
                ), (
                    row_cols_v[idx + count_in_a_row - 1],
                    (r, c_end),
                )
                end = self.to_spreadsheet(r, c_end)
                running_str += f" {start} ... {end}"
                idx += count_in_a_row
            else:
                running_str += " " + self.to_spreadsheet(r, c)
                idx += 1
        return running_str

    def find_contiguous_shapes(self, grid: Grid, color: int) -> List[np.ndarray]:
        labeled_array, num_features = label(grid == color)
        shapes = []
        for i in range(1, num_features + 1):
            shapes.append(np.argwhere(labeled_array == i))
        if self.sort_by_count:
            shapes = sorted(shapes, key=lambda x: len(x), reverse=True)
        return shapes

    def find_contiguous_shapes_moore(self, grid: Grid, color: int) -> List[np.ndarray]:
        s = generate_binary_structure(2, 2)
        labeled_array, num_features = label(grid == color, structure=s)
        shapes = []
        for i in range(1, num_features + 1):
            shapes.append(np.argwhere(labeled_array == i))
        if self.sort_by_count:
            shapes = sorted(shapes, key=lambda x: len(x), reverse=True)
        return shapes

    def encode(self, grid: Grid, connected_component: Optional[int] = None) -> str:
        out = "[["
        if connected_component is None:
            connected_component = self.connected_component
        if connected_component == 4:
            out += "(4CC)\n"
        elif connected_component == 8:
            out += "(8CC)\n"

        color_shapes = []

        for color in range(10):
            if connected_component == 4:
                contiguous_shapes = self.find_contiguous_shapes(grid, color)
            elif connected_component == 8:
                contiguous_shapes = self.find_contiguous_shapes_moore(grid, color)
            color_shapes.append(contiguous_shapes)

        if self.sort_by_count:
            sorted_index = np.argsort(
                [
                    sum([len(shape) for shape in contiguous_shapes])
                    for contiguous_shapes in color_shapes
                ]
            )
        else:
            sorted_index = np.arange(10)

        for color in sorted_index:
            contiguous_shapes = color_shapes[color]
            if len(contiguous_shapes) == 0:
                continue
            shape_strings = []
            for shape in contiguous_shapes:
                if self.normalized:
                    min_i = min(i for i, j in shape)
                    min_j = min(j for i, j in shape)
                    normalized = [
                        (i - min_i, j - min_j)
                        for i, j in sorted(shape, key=lambda x: (int(x[0]), int(x[1])))
                    ]
                    basic_shape_str = self.to_spreadsheet_with_dots(normalized)
                    if not self.disable_absolute:
                        shape_str = (
                            "[Abs. "
                            + self.to_spreadsheet(min_i, min_j)
                            + "]"
                            + basic_shape_str
                        )
                    else:
                        shape_str = basic_shape_str
                else:
                    shape = [
                        (i, j)
                        for i, j in sorted(shape, key=lambda x: (int(x[0]), int(x[1])))
                    ]
                    shape_str = self.to_spreadsheet_with_dots(shape)
                shape_strings.append(shape_str)

            full_str = " | ".join(shape_strings)
            if self.max_token_per_color and self.max_token_per_color < len(
                full_str.split(" ")
            ):
                color_str = " [OMITTED DUE TO EXCESSIVE LENGTH]"
            else:
                color_str = full_str

            out += f"{color}: {color_str}\n"

        return out + "]]"

    def parse_position(self, pos):
        # find the letter part
        letter = re.findall(r"[A-Z]+", pos)[0]
        # find the number part
        number = re.findall(r"[0-9]+", pos)[0]
        row = int(number) - 1
        column = self.spreadsheet_col_labels.index(letter)
        return row, column

    def decode(self, encoded_str: str) -> Grid:
        encoded_str = encoded_str.replace("[[", "").replace("]]", "")
        encoded_str = encoded_str.replace("(8CC)\n", "").replace("(4CC)\n", "")
        max_row, max_col = (
            30,
            30,
        )  # Adjusted for the given example, can be adjusted if needed
        grid = np.full(
            (max_row, max_col), -1
        )  # Initialize with -1 to indicate empty cells

        # Process each color and its components
        for line in encoded_str.strip().split("\n"):
            color, components = line.split(": ")
            color = int(color)
            components = components.split(" | ")
            for component in components:
                component = component.replace("[Abs. ", "").replace("]", "")
                component = component.replace(" ... ", "...")
                abs_pos, *rel_positions = component.split()
                abs_pos = self.parse_position(abs_pos)
                abs_row, abs_col = abs_pos
                for rel_position in rel_positions:
                    if "..." in rel_position:
                        start_pos, end_pos = rel_position.split("...")
                        start_row, start_col = self.parse_position(start_pos)
                        end_row, end_col = self.parse_position(end_pos)
                        for row in range(start_row, end_row + 1):
                            grid[abs_row + row][
                                abs_col + start_col : abs_col + end_col + 1
                            ] = color
                    else:
                        rel_row, rel_col = self.parse_position(rel_position)
                        grid[abs_row + rel_row][abs_col + rel_col] = color

        # crop the grid from -1s
        rows = np.all(grid == -1, axis=1)
        cols = np.all(grid == -1, axis=0)
        grid = grid[~rows]
        grid = grid[:, ~cols]
        # replace remaining -1s with 0
        grid = np.where(grid == -1, 0, grid)
        return grid


class CompositeRepresenter(GridRepresenter):
    connected_component: int = 4

    def __init__(self, representers: List[GridRepresenter]):
        self.representers = representers

    def encode(
        self,
        grid: Grid,
        actives: Optional[List[int]] = None,
        connected_component: Optional[int] = None,
    ) -> str:
        if actives is not None:
            representers = [self.representers[i] for i in actives]
        else:
            representers = self.representers

        out = ""
        for representer in representers:
            if isinstance(representer, ConnectedComponentRepresenter):
                kwargs = {"connected_component": connected_component}
            else:
                kwargs = {}
            out += representer.encode(grid, **kwargs)
            out += "\n"
        return out.strip()

    def decode(self, encoded_str: str, actives: Optional[List[int]] = None) -> Grid:
        # Decoding logic for CompositeRepresenter is complex and depends on the specific encoding format.
        # This is a placeholder for the actual implementation.
        raise NotImplementedError(
            "Decoding for CompositeRepresenter is not implemented."
        )


class ImageGridRepresenter(GridRepresenter):
    cmap: ListedColormap = ListedColormap(
        [
            "#000000",
            "#0074D9",
            "#FF4136",
            "#2ECC40",
            "#FFDC00",
            "#AAAAAA",
            "#F012BE",
            "#FF851B",
            "#7FDBFF",
            "#870C25",
        ]
    )

    cnames: List[str] = [
        "black",
        "blue",
        "red",
        "green",
        "yellow",
        "gray",
        "magenta",
        "orange",
        "lightblue",
        "brown",
    ]

    @np_lru_cache(maxsize=8096)
    def encode(self, grid: Grid) -> str:
        # make sure the actual pixels are based on the grid's size
        fig, ax = plt.subplots(figsize=(len(grid[0]) / 2, len(grid) / 2))
        norm = Normalize(vmin=0, vmax=9)
        ax.imshow(grid, cmap=self.cmap, norm=norm)
        ax.grid(True, which="both", color="lightgrey", linewidth=0.5)
        ax.set_yticks([x - 0.5 for x in range(1 + len(grid))])
        ax.set_xticks([x - 0.5 for x in range(1 + len(grid[0]))])
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        plt.tight_layout()
        # Save the plot to a BytesIO object
        buf = BytesIO()
        plt.savefig(buf, format="png", bbox_inches="tight")
        plt.close(fig)
        buf.seek(0)

        # Encode the image in base64
        img_base64 = base64.b64encode(buf.read()).decode("utf-8")
        return img_base64

    def display(self, encoded_str: str):
        imgdata = base64.b64decode(encoded_str)
        plt.imshow(plt.imread(BytesIO(imgdata)))
        plt.axis("off")
        plt.show()

    def decode(self, encoded_str: str, **kwargs) -> Grid:
        raise NotImplementedError(
            "Decoding for ImageGridRepresenter is not implemented."
        )


class ConnectedComponentRepresenterV2(GridRepresenter):
    def __init__(self, sort_by_count: bool = False, connected_component: int = 4):
        self.sort_by_count = sort_by_count
        self.connected_component = connected_component

    def find_contiguous_shapes(
        self, grid: Grid, color: int, include_diagonals=False
    ) -> List[np.ndarray]:
        if include_diagonals:
            mask = generate_binary_structure(2, 2)
        else:
            mask = None
        labeled_array, num_features = label(grid == color, structure=mask)
        shapes = []
        for i in range(1, num_features + 1):
            shapes.append(np.argwhere(labeled_array == i))
        if self.sort_by_count:
            shapes = sorted(shapes, key=lambda x: len(x), reverse=True)
        return shapes

    def encode(self, grid: Grid, connected_component: Optional[int] = None) -> str:
        if connected_component is None:
            connected_component = self.connected_component

        color_shapes = []

        for color in range(10):
            contiguous_shapes = self.find_contiguous_shapes(
                grid, color, include_diagonals=connected_component == 8
            )
            color_shapes.append(contiguous_shapes)

        if self.sort_by_count:
            sorted_index = np.argsort(
                [
                    sum([len(shape) for shape in contiguous_shapes])
                    for contiguous_shapes in color_shapes
                ]
            )
        else:
            sorted_index = np.arange(10)
        # specify the shape
        output = f"(height={grid.shape[0]}, width={grid.shape[1]})\n"
        for k, color in enumerate(sorted_index):
            # skip color 0
            if color == 0:
                continue
            contiguous_shapes = color_shapes[color]
            if len(contiguous_shapes) == 0:
                continue
            shape_strings = []
            for shape in contiguous_shapes:
                min_i, min_j = np.min(shape, axis=0)
                max_i, max_j = np.max(shape, axis=0)

                subshape = grid[min_i : max_i + 1, min_j : max_j + 1]
                subshape_str = PythonListGridRepresenter().encode(subshape)

                shape_str = (
                    f"Shape(color={color}, pos=({min_i},{min_j}), grid={subshape_str})"
                )
                shape_strings.append(shape_str)

            output += "- " + "\n- ".join(shape_strings)
            if k != len(sorted_index) - 1:
                output += "\n"
        return output

    def decode(self, encoded_str: str) -> Grid:
        return None

    def __repr__(self) -> str:
        return f"ConnectedComponentRepresenterV2(sort_by_count={self.sort_by_count}, connected_component={self.connected_component})"


# =============== Example REPRESENTATION ===============


class TextExampleRepresenter(ExampleRepresenter):
    def __init__(
        self,
        io_sep: str = " -> ",
        input_header: str = "",
        output_header: str = "",
        output_footer="",
        grid_representer: GridRepresenter = PythonListGridRepresenter(),
    ):
        self.io_sep = io_sep
        self.input_header = input_header
        self.output_header = output_header
        self.output_footer = output_footer
        self.grid_representer = grid_representer

    def encode(self, example: Example, **kwargs) -> Tuple[str, str]:
        input_str = self.grid_representer.encode(example.input, **kwargs)
        if self.input_header:
            input_header = self.input_header
        else:
            input_header = ""

        output_str = self.grid_representer.encode(example.output, **kwargs)
        if self.output_header:
            output_header = self.output_header
        else:
            output_header = ""

        return (
            f"{input_header}{input_str}{self.io_sep}{output_header}",
            f"{output_str}{self.output_footer}",
        )

    def decode(self, encoded: Tuple[str, str], **kwargs) -> Example:
        input_str, output_str = encoded
        input_str = input_str.replace(self.input_header, "").replace(
            self.output_header, ""
        )
        if self.io_sep != "\n":
            input_str = input_str.replace(self.io_sep, "")
            input_str.strip()

        output_str = (
            output_str.replace(self.input_header, "")
            .replace(self.output_header, "")
            .strip()
        )

        input_grid = self.grid_representer.decode(input_str, **kwargs)
        output_grid = self.grid_representer.decode(output_str, **kwargs)
        return Example(input=input_grid, output=output_grid)

    def __repr__(self) -> str:
        return f"TextExampleRepresenter(io_sep={self.io_sep!r}, input_header={self.input_header!r}, output_header={self.output_header!r}, output_footer={self.output_footer!r}, grid_representer={repr(self.grid_representer)})"

    def __str__(self) -> str:
        return repr(self)


class ImageExampleRepresenter(ExampleRepresenter):
    def __init__(self, grid_representer: ImageGridRepresenter = ImageGridRepresenter()):
        self.grid_representer = grid_representer

    def encode(self, example: Example, **kwargs) -> str:
        input_grid = example.input
        output_grid = example.output

        # Create a figure with two subplots side by side
        # max height
        height = max(len(input_grid), len(output_grid))
        # max width
        width = max(len(input_grid[0]), len(output_grid[0]))
        fig, axes = plt.subplots(1, 2, figsize=(height, width))

        # Plot input grid
        grid = input_grid
        norm = Normalize(vmin=0, vmax=9)
        ax = axes[0]
        ax.imshow(input_grid, cmap=self.grid_representer.cmap, norm=norm)
        ax.set_title("Input")
        ax.grid(True, which="both", color="lightgrey", linewidth=0.5)
        ax.set_yticks([x - 0.5 for x in range(1 + len(grid))])
        ax.set_xticks([x - 0.5 for x in range(1 + len(grid[0]))])
        ax.set_xticklabels([])
        ax.set_yticklabels([])

        # Plot output grid
        grid = output_grid
        ax = axes[1]
        ax.imshow(output_grid, cmap=self.grid_representer.cmap, norm=norm)
        ax.set_title("Output")
        ax.grid(True, which="both", color="lightgrey", linewidth=0.5)
        ax.set_yticks([x - 0.5 for x in range(1 + len(grid))])
        ax.set_xticks([x - 0.5 for x in range(1 + len(grid[0]))])
        ax.set_xticklabels([])
        ax.set_yticklabels([])

        plt.tight_layout()

        # Save the plot to a BytesIO object
        buf = BytesIO()
        plt.savefig(buf, format="png", bbox_inches="tight")
        plt.close(fig)
        buf.seek(0)

        # Encode the image in base64
        img_base64 = base64.b64encode(buf.read()).decode("utf-8")
        return img_base64

    def display(self, encoded_str: str):
        imgdata = base64.b64decode(encoded_str)
        plt.imshow(plt.imread(BytesIO(imgdata)))
        plt.axis("off")
        plt.show()

    def decode(self, encoded_str: str, **kwargs) -> Example:
        raise NotImplementedError(
            "Decoding for ImageExampleRepresenter is not implemented."
        )


class DiffExampleRepresenter(ExampleRepresenter):
    def __init__(
        self,
        grid_representer: GridRepresenter = PythonListGridRepresenter(),
        io_sep: str = "\n\n",
        input_header: str = "INPUT:\n",
        output_header: str = "OUTPUT:\n",
        output_footer: str = "",
        use_output: bool = True,
        diff_sep: str = " + ",
        diff_output_sep: str = ",",
    ):
        self.io_sep = io_sep
        self.input_header = input_header
        self.output_header = output_header
        self.diff_output_sep = diff_output_sep
        self.diff_sep = diff_sep
        self.use_output = use_output
        self.output_footer = output_footer
        self.grid_representer = grid_representer

    def encode(self, example: Example, **kwargs) -> Tuple[str, str]:

        if np.shape(example.input) != np.shape(example.output):
            return None, None
        diff_str = self.grid_representer.encode(example.output - example.input)
        diff_str = diff_str.replace("  ", " ").replace("[[ ", "[[").replace("[ ", "[")
        input_str = self.grid_representer.encode(example.input, **kwargs)

        if self.use_output:
            output_str = self.grid_representer.encode(example.output, **kwargs)
            return (
                f"{self.input_header}{input_str}{self.io_sep}{self.output_header}",
                f"{self.diff_sep}{diff_str}{self.diff_output_sep}{output_str}{self.output_footer}",
            )
        else:
            return (
                f"{self.input_header}{input_str}{self.io_sep}{self.output_header}",
                f"{self.diff_sep}{diff_str}{self.output_footer}",
            )

    def decode(self, encoded: Tuple[str, str], **kwargs) -> Example:
        input_str, output_str = encoded
        input_str = (
            input_str.replace(self.input_header, "")
            .replace(self.output_header, "")
            .replace(self.io_sep, "")
            .strip()
        )
        output_str = (
            output_str.replace(self.diff_sep.strip(), "")
            .replace(self.output_footer, "")
            .strip()
        )
        input_grid = self.grid_representer.decode(input_str, **kwargs)
        if self.use_output:
            diff_str, output_str = output_str.split(self.diff_output_sep)
            output_grid = self.grid_representer.decode(output_str, **kwargs)
        else:
            diff_grid = self.grid_representer.decode(output_str, **kwargs)
            output_grid = input_grid + diff_grid
        return Example(input=input_grid, output=output_grid)

    def __repr__(self) -> str:
        return f"DiffExampleRepresenter(io_sep={self.io_sep!r}, input_header={self.input_header!r}, output_header={self.output_header!r}, output_footer={self.output_footer!r}, use_output={self.use_output}, diff_sep={self.diff_sep!r}, diff_output_sep={self.diff_output_sep!r}, grid_representer={repr(self.grid_representer)})"

    def __str__(self) -> str:
        return repr(self)


# =============== TASK REPRESENTATION ===============


class TextTaskRepresenter(TaskRepresenter):
    def __init__(
        self,
        train_header: str = "==TRAIN==\n",
        train_test_sep: str = "\n\n",
        test_header: str = "==TEST==\n",
        example_sep: str = "\n\n",
        example_representer: ExampleRepresenter = TextExampleRepresenter(),
    ):
        self.train_header = train_header
        self.train_test_sep = train_test_sep
        self.test_header = test_header
        self.example_sep = example_sep
        self.example_representer = example_representer

    def encode(self, task: Task, **kwargs) -> Tuple[str, str, str]:
        trains = self.train_header
        for train_example in task.train_examples:
            query, output = self.example_representer.encode(train_example, **kwargs)
            trains += query + output
            trains += self.example_sep

        trains = trains[: -len(self.example_sep)]

        demonstrations = trains

        test = self.test_header
        query, output = self.example_representer.encode(task.test_example, **kwargs)
        test += query

        return demonstrations, test, output

    def decode(self, encoded: Tuple[str, str, str], **kwargs) -> Task:
        train_examples = []
        demonstrations, test, encoded_output = encoded

        train_str = demonstrations.replace(self.train_header, "").strip()
        for example_str in train_str.split(
            self.example_sep + self.example_representer.input_header
        ):
            input_str, output_str = example_str.split(
                self.example_representer.io_sep + self.example_representer.output_header
            )
            train_example = self.example_representer.decode(
                (input_str, output_str), **kwargs
            )
            train_examples.append(train_example)

        test_input_str = test.replace(self.test_header, "")
        test_example = self.example_representer.decode(
            (test_input_str, encoded_output), **kwargs
        )

        return Task(train_examples=train_examples, test_example=test_example)

    def __repr__(self) -> str:
        return f"TextTaskRepresenter(train_header={self.train_header!r}, train_test_sep={self.train_test_sep!r}, test_header={self.test_header!r}, example_sep={self.example_sep!r}, example_representer={repr(self.example_representer)})"

    def __str__(self) -> str:
        return repr(self)


class ImageTaskRepresenter(TaskRepresenter):
    example_representer: ImageExampleRepresenter = ImageExampleRepresenter()

    def __init__(
        self, example_representer: Optional[ImageExampleRepresenter] = None
    ) -> None:
        if example_representer is not None:
            self.example_representer = example_representer

    def encode(self, task: Task, show_test_output=False, **kwargs) -> str:
        height = task.max_height()
        width = task.max_width()
        examples = task.train_examples + [task.test_example]
        height = len(examples)
        width = 2
        fig, axes = plt.subplots(len(examples), 2, figsize=(3 * height, 3 * width))
        norm = Normalize(vmin=0, vmax=9)
        for k, example in enumerate(examples):
            input_grid = example.input
            output_grid = example.output
            # Plot input grid
            grid = input_grid
            ax = axes[k, 0]
            ax.imshow(
                grid, cmap=self.example_representer.grid_representer.cmap, norm=norm
            )
            ax.set_title("Input")
            ax.grid(True, which="both", color="lightgrey", linewidth=0.5)
            ax.set_yticks([x - 0.5 for x in range(1 + len(grid))])
            ax.set_xticks([x - 0.5 for x in range(1 + len(grid[0]))])
            ax.set_xticklabels([])
            ax.set_yticklabels([])

            # Plot output grid
            grid = output_grid
            ax = axes[k, 1]
            if not show_test_output and k == len(examples) - 1:
                # display black image
                grid = np.zeros((1, 1))
                ax.imshow(
                    grid, cmap=self.example_representer.grid_representer.cmap, norm=norm
                )
                ax.set_title("(Hidden)")
                ax.grid(True, which="both", color="lightgrey", linewidth=0.5)
                ax.set_yticks([x - 0.5 for x in range(1 + len(grid))])
                ax.set_xticks([x - 0.5 for x in range(1 + len(grid[0]))])
                ax.set_xticklabels([])
                ax.set_yticklabels([])
            else:
                ax = axes[k, 1]
                ax.imshow(
                    grid, cmap=self.example_representer.grid_representer.cmap, norm=norm
                )
                ax.set_title("Output")
                ax.grid(True, which="both", color="lightgrey", linewidth=0.5)
                ax.set_yticks([x - 0.5 for x in range(1 + len(grid))])
                ax.set_xticks([x - 0.5 for x in range(1 + len(grid[0]))])
                ax.set_xticklabels([])
                ax.set_yticklabels([])

        plt.tight_layout()

        # Save the plot to a BytesIO object
        buf = BytesIO()
        plt.savefig(buf, format="png", bbox_inches="tight")
        plt.close(fig)
        buf.seek(0)

        # Encode the image in base64
        img_base64 = base64.b64encode(buf.read()).decode("utf-8")
        return img_base64

    def display(self, encoded_str: str):
        imgdata = base64.b64decode(encoded_str)
        plt.imshow(plt.imread(BytesIO(imgdata)))
        plt.axis("off")
        plt.show()

    def decode(self, encoded_str: str, **kwargs) -> Example:
        raise NotImplementedError(
            "Decoding for ImageExampleRepresenter is not implemented."
        )


if __name__ == "__main__":

    example_representer = TextExampleRepresenter(
        grid_representer=WordGridRepresenter(),
        input_header="Input:\n",
        output_header="\nOutput:\n",
        io_sep="\n",
    )

    grid = np.array([[1, 1, 1], [0, 0, 0], [1, 1, 1]])

    example = Example(input=grid, output=grid)
    print(example_representer.encode(example))
    assert example == example_representer.decode(example_representer.encode(example))

    grid = np.array([[1, 1, 1], [0, 0, 0], [1, 1, 1]])
    representer = ConnectedComponentRepresenter()
    print(representer.encode(grid))
    assert np.array_equal(grid, representer.decode(representer.encode(grid)))
    representer.connected_component = 8
    print(representer.encode(grid))
    assert np.array_equal(grid, representer.decode(representer.encode(grid)))
    representer = DelimitedGridRepresenter()
    print(representer.encode(grid))
    assert np.array_equal(grid, representer.decode(representer.encode(grid)))
    representer = PythonListGridRepresenter()
    print(representer.encode(grid))
    assert np.array_equal(grid, representer.decode(representer.encode(grid)))
    representer = CompositeRepresenter(
        [DelimitedGridRepresenter(), ConnectedComponentRepresenter()]
    )
    print(representer.encode(grid))
    example = Example(input=grid, output=grid)
    representer = TextExampleRepresenter()
    print(representer.encode(example))
    assert example == representer.decode(representer.encode(example))
    task = Task(train_examples=[example, example], test_example=example)
    representer = TextTaskRepresenter()
    assert task == representer.decode(representer.encode(task))
    print(representer.encode(task))
    print(representer.display(representer.encode(task)))

    # image
    representer = ImageGridRepresenter()
    print(representer.encode(grid))
    # save as png
    base64_img = representer.encode(grid)
    with open("output.png", "wb") as f:
        f.write(base64.b64decode(base64_img))

    representer = ImageExampleRepresenter()
    print(representer.encode(example))
    base64_img = representer.encode(Example(input=grid, output=grid))
    with open("example.png", "wb") as f:
        f.write(base64.b64decode(base64_img))

    representer = ImageTaskRepresenter()
    print(representer.encode(task))
    base64_img = representer.encode(task)
    with open("task.png", "wb") as f:
        f.write(base64.b64decode(base64_img))

    grid = np.array([[1, 1, 1], [0, 0, 0], [1, 1, 1]])
    representer = ConnectedComponentRepresenterV2()
    print(representer.encode(grid))

    representer = DiffExampleRepresenter(use_output=False)
    example = Example(input=grid, output=2 * grid)
    assert example == representer.decode(representer.encode(example))
    representer_str = repr(representer)
    print(representer_str)
